import asyncio
import functools
import inspect
import itertools
import logging
import os
import re
import shlex
import shutil
import sys
from datetime import timedelta
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import click
import humanize
from click.types import convert_type
from yarl import URL

from neuromation.api import Action, Client, JobStatus, TagOption, Volume
from neuromation.api.url_utils import uri_from_cli

from .parse_utils import parse_timedelta
from .root import Root
from .stats import upload_gmp_stats
from .version_utils import run_version_checker


log = logging.getLogger(__name__)

_T = TypeVar("_T")

DEPRECATED_HELP_NOTICE = " " + click.style("(DEPRECATED)", fg="red")
DEPRECATED_INVOKE_NOTICE = "DeprecationWarning: The command {name} is deprecated."


async def _run_async_function(
    init_client: bool,
    func: Callable[..., Awaitable[_T]],
    root: Root,
    *args: Any,
    **kwargs: Any,
) -> _T:
    loop = asyncio.get_event_loop()

    if init_client:
        await root.init_client()

        pypi_task: "asyncio.Task[None]" = loop.create_task(
            run_version_checker(root.client, root.disable_pypi_version_check)
        )
        stats_task: "asyncio.Task[None]" = loop.create_task(
            upload_gmp_stats(
                root.client, root.command_path, root.command_params, root.skip_gmp_stats
            )
        )
    else:
        pypi_task = loop.create_task(asyncio.sleep(0))  # do nothing
        stats_task = loop.create_task(asyncio.sleep(0))  # do nothing

    try:
        return await func(root, *args, **kwargs)
    finally:
        stats_task.cancel()
        try:
            await stats_task
        except asyncio.CancelledError:
            pass
        except Exception:
            log.debug("Usage stats sending has failed", exc_info=True)
        pypi_task.cancel()
        try:
            await pypi_task
        except asyncio.CancelledError:
            pass
        except Exception:
            log.debug("PyPI checker has failed", exc_info=True)


def _wrap_async_callback(
    callback: Callable[..., Awaitable[_T]],
    init_client: bool = True,
) -> Callable[..., _T]:
    assert inspect.iscoroutinefunction(callback)

    # N.B. the decorator implies @click.pass_obj
    @click.pass_obj
    @functools.wraps(callback)
    def wrapper(root: Root, *args: Any, **kwargs: Any) -> _T:
        return root.run(
            _run_async_function(init_client, callback, root, *args, **kwargs),
        )

    return wrapper


class HelpFormatter(click.HelpFormatter):
    def write_usage(self, prog: str, args: str = "", prefix: str = "Usage:") -> None:
        super().write_usage(prog, args, prefix=click.style(prefix, bold=True) + " ")

    def write_heading(self, heading: str) -> None:
        self.write(
            click.style(
                "%*s%s:\n" % (self.current_indent, "", heading),
                bold=True,
                underline=False,
            )
        )


class Context(click.Context):
    def make_formatter(self) -> click.HelpFormatter:
        return HelpFormatter(
            width=self.terminal_width, max_width=self.max_content_width
        )


def split_examples(help: str) -> List[str]:
    return re.split("Example[s]:\n", help, re.IGNORECASE)


def format_example(example: str, formatter: click.HelpFormatter) -> None:
    with formatter.section(click.style("Examples", bold=True, underline=False)):
        for line in example.splitlines():
            is_comment = line.startswith("#")
            if is_comment:
                formatter.write_text("\b\n" + click.style(line, dim=True))
            else:
                formatter.write_text("\b\n" + " ".join(shlex.split(line)))


class NeuroClickMixin:
    def get_params(self, ctx: click.Context) -> List[click.Parameter]:
        # super() is available after using as a mixin
        ret = super().get_params(ctx)  # type: ignore
        args = [i for i in ret if not isinstance(i, click.Option)]
        opts = [i for i in ret if isinstance(i, click.Option)]

        help_names = self.get_help_option_names(ctx)  # type: ignore

        def sort_key(opt: click.Option) -> Tuple[bool, str]:
            flag = set(opt.opts) & help_names or set(opt.secondary_opts) & help_names
            return (not flag, opt.name)

        return args + sorted(opts, key=sort_key)

    def get_help_option(self, ctx: click.Context) -> Optional[click.Option]:
        help_options = self.get_help_option_names(ctx)  # type: ignore
        if not help_options or not self.add_help_option:  # type: ignore
            return None

        def show_help(ctx: click.Context, param: Any, value: Any) -> None:
            if value and not ctx.resilient_parsing:
                print_help(ctx)

        return Option(
            help_options,
            is_flag=True,
            is_eager=True,
            expose_value=False,
            callback=show_help,
            help="Show this message and exit.",
        )

    def get_short_help_str(self, limit: int = 45) -> str:
        text = super().get_short_help_str(limit=limit)  # type: ignore
        if text.endswith(".") and not text.endswith("..."):
            text = text[:-1]
        return text

    def format_help_text(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        """Writes the help text to the formatter if it exists."""
        deprecated = self.deprecated  # type: ignore
        help = self.help  # type: ignore
        if help:
            help_text, *examples = split_examples(help)
            if help_text:
                formatter.write_paragraph()
                with formatter.indentation():
                    if deprecated:
                        help_text += DEPRECATED_HELP_NOTICE
                    formatter.write_text(help_text)
            examples = [example.strip() for example in examples]

            for example in examples:
                format_example(example, formatter)
        elif deprecated:
            formatter.write_paragraph()
            with formatter.indentation():
                formatter.write_text(DEPRECATED_HELP_NOTICE)

    def make_context(
        self,
        info_name: str,
        args: Sequence[str],
        parent: Optional[click.Context] = None,
        **extra: Any,
    ) -> Context:
        for key, value in self.context_settings.items():  # type: ignore
            if key not in extra:
                extra[key] = value
        ctx = Context(self, info_name=info_name, parent=parent, **extra)  # type: ignore
        with ctx.scope(cleanup=False):
            self.parse_args(ctx, args)  # type: ignore
        return ctx


class NeuroGroupMixin(NeuroClickMixin):
    def format_options(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        self.format_commands(ctx, formatter)  # type: ignore


def _collect_params(cmd: click.Command, ctx: click.Context) -> Dict[str, Optional[str]]:
    params = ctx.params.copy()
    for param in cmd.get_params(ctx):
        if param.name not in params:
            continue
        if params[param.name] == param.get_default(ctx):
            # drop default param
            del params[param.name]
            continue
        if param.param_type_name != "option":
            # save name only
            params[param.name] = None
        else:
            if getattr(param, "secure", True):
                params[param.name] = None
            else:
                params[param.name] = str(params[param.name])
    return params


class Command(NeuroClickMixin, click.Command):
    def __init__(
        self,
        callback: Any,
        init_client: bool = True,
        wrap_async: bool = True,
        **kwargs: Any,
    ) -> None:
        if wrap_async:
            callback = _wrap_async_callback(callback, init_client=init_client)
        super().__init__(
            callback=callback,
            **kwargs,
        )
        self.init_client = init_client

    def invoke(self, ctx: click.Context) -> Any:
        """Given a context, this invokes the attached callback (if it exists)
        in the right way.
        """
        if self.deprecated:
            click.echo(
                click.style(DEPRECATED_INVOKE_NOTICE.format(name=self.name), fg="red"),
                err=True,
            )
        if self.callback is not None:
            # Collect arguments for sending to google analytics
            ctx2 = ctx
            params = [_collect_params(ctx2.command, ctx2)]
            while ctx2.parent:
                ctx2 = ctx2.parent
                params.append(_collect_params(ctx2.command, ctx2))
            params.reverse()
            root = cast(Root, ctx.obj)
            root.command_path = ctx.command_path
            root.command_params = params
            return ctx.invoke(self.callback, **ctx.params)


def command(
    name: Optional[str] = None, cls: Type[Command] = Command, **kwargs: Any
) -> Command:
    return click.command(name=name, cls=cls, **kwargs)  # type: ignore


class Group(NeuroGroupMixin, click.Group):
    def command(
        self, *args: Any, **kwargs: Any
    ) -> Callable[[Callable[..., Any]], Command]:
        def decorator(f: Callable[..., Any]) -> Command:
            cmd = command(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator

    def group(
        self, *args: Any, **kwargs: Any
    ) -> Callable[[Callable[..., Any]], "Group"]:
        def decorator(f: Callable[..., Any]) -> Group:
            cmd = group(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator

    def list_commands(self, ctx: click.Context) -> Iterable[str]:
        return self.commands

    def invoke(self, ctx: click.Context) -> None:
        if not ctx.args and not ctx.protected_args:
            print_help(ctx)
        else:
            super().invoke(ctx)


def group(name: Optional[str] = None, **kwargs: Any) -> Group:
    kwargs.setdefault("cls", Group)
    kwargs.setdefault("invoke_without_command", True)
    return click.group(name=name, **kwargs)  # type: ignore


def print_help(ctx: click.Context) -> None:
    root = cast(Root, ctx.obj)
    if root is None:
        tty = all(f.isatty() for f in [sys.stdin, sys.stdout, sys.stderr])
        terminal_size = shutil.get_terminal_size()
    else:
        tty = root.tty
        terminal_size = root.terminal_size

    pager_maybe(ctx.get_help().splitlines(), tty, terminal_size)
    ctx.exit()


class DeprecatedGroup(NeuroGroupMixin, click.MultiCommand):
    def __init__(
        self, origin: click.MultiCommand, name: Optional[str] = None, **attrs: Any
    ) -> None:
        attrs.setdefault("help", f"Alias for {origin.name}")
        attrs.setdefault("deprecated", True)
        super().__init__(name, **attrs)
        self.origin = origin

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        return self.origin.get_command(ctx, cmd_name)

    def list_commands(self, ctx: click.Context) -> Iterable[str]:
        return self.origin.list_commands(ctx)


def alias(
    origin: click.Command,
    name: str,
    *,
    deprecated: bool = True,
    hidden: Optional[bool] = None,
    help: Optional[str] = None,
) -> click.Command:
    if help is None:
        help = f"Alias for {origin.name}."
    if hidden is None:
        hidden = origin.hidden

    return Command(
        name=name,
        context_settings=origin.context_settings,
        callback=origin.callback,
        params=origin.params,
        help=help,
        epilog=origin.epilog,
        short_help=origin.short_help,
        options_metavar=origin.options_metavar,
        add_help_option=origin.add_help_option,
        hidden=hidden,
        deprecated=deprecated,
        wrap_async=False,
    )


class Option(click.Option):
    def __init__(self, *args: Any, secure: bool = False, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.secure = secure


def option(*param_decls: Any, **attrs: Any) -> Callable[..., Any]:
    option_attrs = attrs.copy()
    option_attrs.setdefault("cls", Option)
    typ = convert_type(attrs.get("type"), attrs.get("default"))
    autocompletion = getattr(typ, "complete", None)
    option_attrs.setdefault("autocompletion", autocompletion)
    return click.option(*param_decls, **option_attrs)


def argument(*param_decls: Any, **attrs: Any) -> Callable[..., Any]:
    arg_attrs = attrs.copy()
    typ = convert_type(attrs.get("type"), attrs.get("default"))
    autocompletion = getattr(typ, "complete", None)
    arg_attrs.setdefault("autocompletion", autocompletion)
    return click.argument(*param_decls, **arg_attrs)


def volume_to_verbose_str(volume: Volume) -> str:
    return (
        f"'{volume.storage_uri}' mounted to '{volume.container_path}' "
        f"in {('ro' if volume.read_only else 'rw')} mode"
    )


JOB_ID_PATTERN = r"job-[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}"


async def resolve_job(
    id_or_name_or_uri: str, *, client: Client, status: Set[JobStatus]
) -> str:
    default_user = client.username
    default_cluster = client.cluster_name
    if id_or_name_or_uri.startswith("job:"):
        uri = uri_from_cli(
            id_or_name_or_uri,
            username=default_user,
            cluster_name=default_cluster,
            allowed_schemes=("job",),
        )
        if uri.host != default_cluster:
            raise ValueError(f"Invalid job URI: cluster_name != '{default_cluster}'")
        owner, _, id_or_name = uri.path.lstrip("/").partition("/")
        if not owner:
            raise ValueError(f"Invalid job URI: missing owner")
        if not id_or_name:
            raise ValueError(
                f"Invalid job URI: owner='{owner}', missing job-id or job-name"
            )
    else:
        id_or_name = id_or_name_or_uri
        owner = default_user

    # Temporary fast path.
    if re.fullmatch(JOB_ID_PATTERN, id_or_name):
        return id_or_name

    try:
        async for job in client.jobs.list(
            name=id_or_name, owners={owner}, reverse=True, limit=1
        ):
            log.debug(f"Job name '{id_or_name}' resolved to job ID '{job.id}'")
            return job.id
    except asyncio.CancelledError:
        raise
    except Exception as e:
        log.error(
            f"Failed to resolve job-name {id_or_name_or_uri} resolved as "
            f"name={id_or_name}, owner={owner} to a job-ID: {e}"
        )

    if owner != default_user:
        raise ValueError(f"Failed to resolve job {id_or_name_or_uri}")
    return id_or_name


SHARE_SCHEMES = ("storage", "image", "job", "blob", "role", "secret", "disk")


def parse_resource_for_sharing(uri: str, root: Root) -> URL:
    """Parses the neuromation resource URI string.
    Available schemes: storage, image, job. For image URIs, tags are not allowed.
    """
    if uri.startswith("image:"):
        image = root.client.parse.remote_image(uri, tag_option=TagOption.DENY)
        uri = str(image)

    uri_res = uri_from_cli(
        uri,
        root.client.username,
        root.client.cluster_name,
        allowed_schemes=SHARE_SCHEMES,
    )
    # URI's for object storage can only operate on bucket level
    if uri_res.scheme == "blob" and "/" in uri_res.path.strip("/"):
        raise ValueError("Only bucket level permissions are supported for Blob Storage")
    return uri_res


def parse_file_resource(uri: str, root: Root) -> URL:
    """Parses the neuromation resource URI string.
    Available schemes: file, storage.
    """
    return uri_from_cli(
        uri,
        root.client.username,
        root.client.cluster_name,
        allowed_schemes=("file", "storage"),
    )


def parse_blob_resource(uri: str, root: Root) -> URL:
    # Username will not be used, just part of the signature
    return uri_from_cli(
        uri, root.client.username, root.client.cluster_name, allowed_schemes=("blob",)
    )


def parse_blob_or_file_resource(uri: str, root: Root) -> URL:
    # Username will not be used, just part of the signature
    return uri_from_cli(
        uri,
        root.client.username,
        root.client.cluster_name,
        allowed_schemes=("blob", "file"),
    )


def parse_secret_resource(uri: str, root: Root) -> URL:
    return uri_from_cli(
        uri,
        root.client.username,
        root.client.cluster_name,
        allowed_schemes=("secret"),
    )


def parse_permission_action(action: str) -> Action:
    try:
        return Action[action.upper()]
    except KeyError:
        valid_actions = ", ".join([a.value for a in Action])
        raise ValueError(
            f"invalid permission action '{action}', allowed values: {valid_actions}"
        )


def do_deprecated_quiet(
    ctx: click.Context, param: Union[click.Option, click.Parameter], value: Any
) -> Any:
    if value and not ctx.obj.quiet:
        ctx.obj.verbosity = -2
        click.echo(
            click.style(
                "DeprecationWarning: "
                "The local option --quiet is deprecated. "
                "Use global option --quiet instead.",
                fg="red",
            ),
            err=True,
        )
        # Patch the logger as it was set up with verbosity=-2.
        root_logger = logging.getLogger()
        handler = root_logger.handlers[-1]
        assert handler.formatter
        format_class = type(handler.formatter)
        handler.setFormatter(format_class())
        handler.setLevel(logging.ERROR)


deprecated_quiet_option: Any = option(
    "-q",
    "--quiet",
    is_flag=True,
    callback=do_deprecated_quiet,
    help="Run command in quiet mode (DEPRECATED)",
    expose_value=False,
    is_eager=True,
)

if sys.version_info >= (3, 7):  # pragma: no cover
    from contextlib import AsyncExitStack
else:
    from async_exit_stack import AsyncExitStack  # noqa


def format_size(value: float) -> str:
    return humanize.naturalsize(value, gnu=True)


def pager_maybe(
    lines: Iterable[str], tty: bool, terminal_size: Tuple[int, int]
) -> None:
    if not tty:
        for line in lines:
            click.echo(line)
        return

    # Enforce ANSI sequence handling (colors etc.)
    os.environ["LESS"] = "-R"

    lines_it: Iterator[str] = iter(lines)
    count = int(terminal_size[1] * 2 / 3)
    handled = list(itertools.islice(lines_it, count))
    if len(handled) < count:
        # lines list is short, just print it
        for line in handled:
            click.echo(line)
    else:
        click.echo_via_pager(
            itertools.chain(["\n".join(handled)], (f"\n{line}" for line in lines_it))
        )


async def calc_life_span(
    client: Client, value: Optional[str], default: str, config_section: str
) -> Optional[float]:
    async def _calc_default_life_span(client: Client) -> timedelta:
        config = await client.config.get_user_config()
        section = config.get(config_section)
        life_span = default
        if section is not None:
            value = section.get("life-span")
            if value is not None:
                life_span = value
        return parse_timedelta(life_span)

    delta = (
        parse_timedelta(value)
        if value is not None
        else await _calc_default_life_span(client)
    )
    seconds = delta.total_seconds()
    if seconds == 0:
        click.secho(
            "Zero job's life-span (--life-span=0) is deprecated "
            "and will be removed in the future neuro CLI release,"
            "use a positive value to avoid resource leakage",
            fg="yellow",
        )
        return None
    assert seconds > 0
    return seconds



from logging import ERROR, WARNING, LogRecord

from neuromation.cli.log_formatter import ConsoleWarningFormatter


formatter = ConsoleWarningFormatter("%(name)s.%(funcName)s: %(message)s")


def test_warning() -> None:
    record = LogRecord("n", WARNING, "p", 1, "warn-message", (), None)
    formatted = formatter.format(record)
    # yellow WARNING
    assert formatted.startswith("\x1b[33mWARNING\x1b[0m")
    # message inside
    assert formatted.find("warn-message") >= 0


def test_error() -> None:
    record = LogRecord("n", ERROR, "p", 1, "error-message", (), None)
    formatted = formatter.format(record)
    # red ERROR
    assert formatted.startswith("\x1b[31mERROR\x1b[0m")
    # message inside
    assert formatted.find("error-message") >= 0



import abc
from types import TracebackType
from typing import Dict, Iterable, Optional, Type

from rich import box
from rich.console import Console, RenderableType
from rich.progress import BarColumn, DownloadColumn, Progress, TaskID, TextColumn
from rich.table import Table

from neuromation.api import (
    AbstractDockerImageProgress,
    ImageCommitFinished,
    ImageCommitStarted,
    ImageProgressPull,
    ImageProgressPush,
    ImageProgressSave,
    ImageProgressStep,
    RemoteImage,
)

from ...api.parsing_utils import Tag
from ..utils import format_size
from .utils import ImageFormatter


class DockerImageProgress(AbstractDockerImageProgress):
    @classmethod
    def create(cls, console: Console, quiet: bool) -> "DockerImageProgress":
        if quiet:
            progress: DockerImageProgress = QuietDockerImageProgress()
        elif console.is_terminal:
            progress = DetailedDockerImageProgress(console)
        else:
            progress = StreamDockerImageProgress(console)
        return progress

    @abc.abstractmethod
    def close(self) -> None:  # pragma: no cover
        pass

    def __enter__(self) -> "DockerImageProgress":
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self.close()

    def _shorten_container_hash(self, container: str) -> str:
        return container[:12]


class QuietDockerImageProgress(DockerImageProgress):
    def pull(self, data: ImageProgressPull) -> None:
        pass

    def push(self, data: ImageProgressPush) -> None:
        pass

    def step(self, data: ImageProgressStep) -> None:
        pass

    def close(self) -> None:
        pass

    def save(self, data: ImageProgressSave) -> None:
        pass

    def commit_started(self, data: ImageCommitStarted) -> None:
        pass

    def commit_finished(self, data: ImageCommitFinished) -> None:
        pass


class DetailedDockerImageProgress(DockerImageProgress):
    def __init__(self, console: Console) -> None:
        self._mapping: Dict[str, TaskID] = {}
        self._progress = Progress(
            TextColumn("[progress.description]{task.fields[layer]}"),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            DownloadColumn(),
            console=console,
            auto_refresh=False,
        )
        self._progress.start()

    def push(self, data: ImageProgressPush) -> None:
        self._progress.log(f"Pushing image [b]{data.src}[/b] => [b]{data.dst}[/b]")

    def pull(self, data: ImageProgressPull) -> None:
        self._progress.log(f"Pulling image [b]{data.src}[/b] => [b]{data.dst}[/b]")

    def step(self, data: ImageProgressStep) -> None:
        if data.layer_id:
            if data.status in ("Layer already exists", "Download complete"):
                current: Optional[float] = 100.0
                total: Optional[float] = 100.0
            else:
                current = float(data.current) if data.current is not None else None
                total = float(data.total) if data.total is not None else None

            layer = data.layer_id
            if layer in self._mapping.keys():
                task = self._mapping[layer]
                self._progress.update(
                    task,
                    description=data.status,
                    completed=current,
                    total=total,
                    refresh=True,
                )
            else:
                task = self._progress.add_task(
                    layer=layer,
                    description=data.status,
                    completed=current or 0,  # type: ignore
                    total=total or 100,  # type: ignore
                )
                self._mapping[layer] = task

        else:
            self._progress.log(data.message)

    def save(self, data: ImageProgressSave) -> None:
        self._progress.log(f"Saving [b]{data.job}[/b] => [b]{data.dst}[/b]")

    def commit_started(self, data: ImageCommitStarted) -> None:
        self._progress.log(
            f"Creating image [b]{data.target_image}[/b] image from the job container",
        )

    def commit_finished(self, data: ImageCommitFinished) -> None:
        self._progress.log("Image created")

    def close(self) -> None:
        self._progress.stop()


class StreamDockerImageProgress(DockerImageProgress):
    def __init__(self, console: Console) -> None:
        self._console = console

    def push(self, data: ImageProgressPush) -> None:
        self._console.print(f"Using local image '{data.src}'")
        self._console.print(f"Using remote image '{data.dst}'")
        self._console.print("Pushing image...")

    def pull(self, data: ImageProgressPull) -> None:
        self._console.print(f"Using remote image '{data.src}'")
        self._console.print(f"Using local image '{data.dst}'")
        self._console.print("Pulling image...")

    def step(self, data: ImageProgressStep) -> None:
        if data.layer_id:
            self._console.print(".", end="")
        else:
            self._console.print(data.message)

    def save(self, data: ImageProgressSave) -> None:
        self._console.print(f"Saving job '{data.job}' to image '{data.dst}'...")

    def commit_started(self, data: ImageCommitStarted) -> None:
        self._console.print(f"Using remote image '{data.target_image}'")
        self._console.print(f"Creating image from the job container...")

    def commit_finished(self, data: ImageCommitFinished) -> None:
        self._console.print("Image created")

    def close(self) -> None:
        pass


class BaseImagesFormatter:
    def __init__(self, image_formatter: ImageFormatter) -> None:
        self._format_image = image_formatter

    @abc.abstractmethod
    def __call__(self, images: Iterable[RemoteImage]) -> RenderableType:
        raise NotImplementedError


class QuietImagesFormatter(BaseImagesFormatter):
    def __call__(self, images: Iterable[RemoteImage]) -> RenderableType:
        table = Table.grid()
        table.add_column("")
        for image in images:
            table.add_row(self._format_image(image))
        return table


class ShortImagesFormatter(BaseImagesFormatter):
    def __call__(self, images: Iterable[RemoteImage]) -> RenderableType:
        table = Table(box=box.SIMPLE_HEAVY)
        table.add_column("Neuro URL", style="bold")
        for image in images:
            table.add_row(self._format_image(image))
        return table


class LongImagesFormatter(BaseImagesFormatter):
    def __call__(self, images: Iterable[RemoteImage]) -> RenderableType:
        table = Table(box=box.SIMPLE_HEAVY)
        table.add_column("Neuro URL", style="bold")
        table.add_column("Docker URL")
        for image in images:
            table.add_row(
                self._format_image(image), image.as_docker_url(with_scheme=True)
            )
        return table


class BaseTagsFormatter:
    def _build_table_for(self) -> Table:
        table = Table(box=box.SIMPLE_HEAVY)
        table.add_column("Tag", style="bold")
        return table

    @abc.abstractmethod
    def __call__(self, image: RemoteImage, tags: Iterable[Tag]) -> RenderableType:
        raise NotImplementedError


class ShortTagsFormatter(BaseTagsFormatter):
    def __call__(self, image: RemoteImage, tags: Iterable[Tag]) -> RenderableType:
        table = self._build_table_for()
        for tag in tags:
            table.add_row(tag.name)
        return table


class LongTagsFormatter(BaseTagsFormatter):
    def __call__(self, image: RemoteImage, tags: Iterable[Tag]) -> RenderableType:
        table = self._build_table_for()
        table.add_column("Size")
        for tag in tags:
            assert tag.name
            assert tag.size
            table.add_row(tag.name, format_size(tag.size))
        return table



import abc
import os
import re
from datetime import datetime, timedelta
from typing import Generic, List, Optional, Sequence, Tuple, TypeVar, Union, cast

import click
from click import BadParameter

from neuromation.api import LocalImage, RemoteImage, TagOption

from .parse_utils import JobColumnInfo, parse_columns, to_megabytes
from .root import Root


# NOTE: these job name defaults are taken from `platform_api` file `validators.py`
JOB_NAME_MIN_LENGTH = 3
JOB_NAME_MAX_LENGTH = 40
JOB_NAME_PATTERN = "^[a-z](?:-?[a-z0-9])*$"
JOB_NAME_REGEX = re.compile(JOB_NAME_PATTERN)
JOB_LIMIT_ENV = "NEURO_CLI_JOB_AUTOCOMPLETE_LIMIT"

_T = TypeVar("_T")


class AsyncType(click.ParamType, Generic[_T], abc.ABC):
    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> _T:
        assert ctx is not None
        root = cast(Root, ctx.obj)
        return root.run(self.async_convert(root, value, param, ctx))

    @abc.abstractmethod
    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> _T:
        pass

    def complete(
        self, ctx: click.Context, args: Sequence[str], incomplete: str
    ) -> List[Tuple[str, Optional[str]]]:
        root = cast(Root, ctx.obj)
        return root.run(self.async_complete(root, ctx, args, incomplete))

    @abc.abstractmethod
    async def async_complete(
        self, root: Root, ctx: click.Context, args: Sequence[str], incomplete: str
    ) -> List[Tuple[str, Optional[str]]]:
        pass


class LocalImageType(click.ParamType):
    name = "local_image"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> LocalImage:
        assert ctx is not None
        root = cast(Root, ctx.obj)
        client = root.run(root.init_client())
        return client.parse.local_image(value)


class ImageType(click.ParamType):
    name = "image"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> RemoteImage:
        assert ctx is not None
        root = cast(Root, ctx.obj)
        client = root.run(root.init_client())
        return client.parse.remote_image(value)


class RemoteTaglessImageType(click.ParamType):
    name = "image"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> RemoteImage:
        assert ctx is not None
        root = cast(Root, ctx.obj)
        client = root.run(root.init_client())
        return client.parse.remote_image(value, tag_option=TagOption.DENY)


class RemoteImageType(click.ParamType):
    name = "image"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> RemoteImage:
        assert ctx is not None
        root = cast(Root, ctx.obj)
        client = root.run(root.init_client())
        return client.parse.remote_image(value)


class LocalRemotePortParamType(click.ParamType):
    name = "local-remote-port-pair"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Tuple[int, int]:
        try:
            local_str, remote_str = value.split(":")
            local, remote = int(local_str), int(remote_str)
            if not (0 < local <= 65535 and 0 < remote <= 65535):
                raise ValueError("Port should be in range 1 to 65535")
            return local, remote
        except ValueError as e:
            raise BadParameter(f"{value} is not a valid port combination: {e}")


LOCAL_REMOTE_PORT = LocalRemotePortParamType()


class MegabyteType(click.ParamType):
    name = "megabyte"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> int:
        return to_megabytes(value)


MEGABYTE = MegabyteType()


class JobNameType(click.ParamType):
    name = "job_name"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> str:
        if (
            len(value) < JOB_NAME_MIN_LENGTH
            or len(value) > JOB_NAME_MAX_LENGTH
            or JOB_NAME_REGEX.match(value) is None
        ):
            raise ValueError(
                f"Invalid job name '{value}'.\n"
                "The name can only contain lowercase letters, numbers and hyphens "
                "with the following rules: \n"
                "  - the first character must be a letter; \n"
                "  - each hyphen must be surrounded by non-hyphen characters; \n"
                f"  - total length must be between {JOB_NAME_MIN_LENGTH} and "
                f"{JOB_NAME_MAX_LENGTH} characters long."
            )
        return value


JOB_NAME = JobNameType()


class JobColumnsType(click.ParamType):
    name = "columns"

    def convert(
        self,
        value: Union[str, List[JobColumnInfo]],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> List[JobColumnInfo]:
        if isinstance(value, list):
            return value
        return parse_columns(value)


JOB_COLUMNS = JobColumnsType()


class PresetType(AsyncType[str]):
    name = "preset"

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        client = await root.init_client()
        if value not in client.presets:
            raise click.BadParameter(
                f"Preset {value} is not valid, "
                "run 'neuro config show' to get a list of available presets",
                ctx,
                param,
            )
        return value

    async def async_complete(
        self, root: Root, ctx: click.Context, args: Sequence[str], incomplete: str
    ) -> List[Tuple[str, Optional[str]]]:
        # async context manager is used to prevent a message about
        # unclosed session
        async with await root.init_client() as client:
            presets = list(client.config.presets)
            return [(p, None) for p in presets if p.startswith(incomplete)]


PRESET = PresetType()


class JobType(AsyncType[str]):
    name = "job"

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_complete(
        self, root: Root, ctx: click.Context, args: Sequence[str], incomplete: str
    ) -> List[Tuple[str, Optional[str]]]:
        async with await root.init_client() as client:
            ret: List[Tuple[str, Optional[str]]] = []
            now = datetime.now()
            limit = int(os.environ.get(JOB_LIMIT_ENV, 100))
            async for job in client.jobs.list(
                since=now - timedelta(days=7), reverse=True, limit=limit
            ):
                job_name = job.name or ""
                for test in (
                    job.id,
                    job_name,
                    f"job:{job.id}",
                    f"job:/{job.owner}/{job.id}",
                    f"job://{job.cluster_name}/{job.owner}/{job.id}",
                ):
                    if test.startswith(incomplete):
                        ret.append((test, job_name))

            return ret


JOB = JobType()



import pathlib

from neuromation.cli.secrets import read_data


def test_read_data_str() -> None:
    assert b"value" == read_data("value")


def test_read_data_file(tmp_path: pathlib.Path) -> None:
    fname = tmp_path / "secret.txt"
    fname.write_bytes(b"file content")
    assert b"file content" == read_data("@" + str(fname))



from dataclasses import replace
from pathlib import Path
from typing import Callable

import toml
from rich.console import RenderableType

from neuromation.api import Client, Cluster, Preset
from neuromation.api.admin import _Quota
from neuromation.api.quota import _QuotaInfo
from neuromation.cli.alias import list_aliases
from neuromation.cli.formatters.config import (
    AliasesFormatter,
    ConfigFormatter,
    QuotaFormatter,
    QuotaInfoFormatter,
)
from neuromation.cli.root import Root


RichCmp = Callable[[RenderableType], None]


class TestConfigFormatter:
    async def test_output(self, root: Root, rich_cmp: RichCmp) -> None:
        out = ConfigFormatter()(root.client.config, {})
        rich_cmp(out)

    async def test_output_for_tpu_presets(
        self,
        make_client: Callable[..., Client],
        cluster_config: Cluster,
        rich_cmp: RichCmp,
    ) -> None:
        presets = dict(cluster_config.presets)

        presets["tpu-small"] = Preset(
            cpu=2,
            memory_mb=2048,
            is_preemptible=False,
            tpu_type="v3-8",
            tpu_software_version="1.14",
        )
        presets["hybrid"] = Preset(
            cpu=4,
            memory_mb=30720,
            is_preemptible=False,
            gpu=2,
            gpu_model="nvidia-tesla-v100",
            tpu_type="v3-64",
            tpu_software_version="1.14",
        )
        new_config = replace(cluster_config, presets=presets)

        client = make_client(
            "https://dev.neu.ro/api/v1", clusters={new_config.name: new_config}
        )
        out = ConfigFormatter()(client.config, {})
        rich_cmp(out)

    async def test_output_with_jobs_available(
        self, root: Root, rich_cmp: RichCmp
    ) -> None:
        available_jobs_counts = {
            "cpu-small": 1,
            "cpu-large": 2,
        }
        out = ConfigFormatter()(root.client.config, available_jobs_counts)
        rich_cmp(out)


bold_start = "\x1b[1m"
bold_end = "\x1b[0m"


class TestQuotaInfoFormatter:
    def test_output(self, rich_cmp: RichCmp) -> None:
        quota = _QuotaInfo(
            cluster_name="default",
            gpu_time_spent=0.0,
            gpu_time_limit=0.0,
            cpu_time_spent=float((9 * 60 + 19) * 60),
            cpu_time_limit=float((9 * 60 + 39) * 60),
        )
        out = QuotaInfoFormatter()(quota)
        rich_cmp(out)

    def test_output_no_quota(self, rich_cmp: RichCmp) -> None:
        quota = _QuotaInfo(
            cluster_name="default",
            gpu_time_spent=0.0,
            gpu_time_limit=float("inf"),
            cpu_time_spent=float((9 * 60 + 19) * 60),
            cpu_time_limit=float("inf"),
        )
        out = QuotaInfoFormatter()(quota)
        rich_cmp(out)

    def test_output_too_many_hours(self, rich_cmp: RichCmp) -> None:
        quota = _QuotaInfo(
            cluster_name="default",
            gpu_time_spent=float((1 * 60 + 29) * 60),
            gpu_time_limit=float((9 * 60 + 59) * 60),
            cpu_time_spent=float((9999 * 60 + 29) * 60),
            cpu_time_limit=float((99999 * 60 + 59) * 60),
        )
        out = QuotaInfoFormatter()(quota)
        rich_cmp(out)

    def test_output_spent_more_than_quota_left_zero(self, rich_cmp: RichCmp) -> None:
        quota = _QuotaInfo(
            cluster_name="default",
            gpu_time_spent=float(9 * 60 * 60),
            gpu_time_limit=float(1 * 60 * 60),
            cpu_time_spent=float(9 * 60 * 60),
            cpu_time_limit=float(2 * 60 * 60),
        )
        out = QuotaInfoFormatter()(quota)
        rich_cmp(out)

    def test_format_time_00h_00m(self) -> None:
        out = QuotaInfoFormatter()._format_time(total_seconds=float(0 * 60))
        assert out == "00h 00m"

    def test_format_time_00h_09m(self) -> None:
        out = QuotaInfoFormatter()._format_time(total_seconds=float(9 * 60))
        assert out == "00h 09m"

    def test_format_time_01h_00m(self) -> None:
        out = QuotaInfoFormatter()._format_time(total_seconds=float(60 * 60))
        assert out == "01h 00m"

    def test_format_time_01h_10m(self) -> None:
        out = QuotaInfoFormatter()._format_time(total_seconds=float(70 * 60))
        assert out == "01h 10m"

    def test_format_time_99h_00m(self) -> None:
        out = QuotaInfoFormatter()._format_time(total_seconds=float(99 * 60 * 60))
        assert out == "99h 00m"

    def test_format_time_99h_59m(self) -> None:
        out = QuotaInfoFormatter()._format_time(
            total_seconds=float((99 * 60 + 59) * 60)
        )
        assert out == "99h 59m"

    def test_format_time_9999h_59m(self) -> None:
        out = QuotaInfoFormatter()._format_time(
            total_seconds=float((9999 * 60 + 59) * 60)
        )
        assert out == "9999h 59m"


class TestQuotaFormatter:
    def test_output(self, rich_cmp: RichCmp) -> None:
        quota = _Quota(
            total_running_jobs=10,
            total_gpu_run_time_minutes=321,
            total_non_gpu_run_time_minutes=123,
        )
        out = QuotaFormatter()(quota)
        rich_cmp(out)

    def test_output_no_quota(self, rich_cmp: RichCmp) -> None:
        quota = _Quota(
            total_running_jobs=None,
            total_non_gpu_run_time_minutes=None,
            total_gpu_run_time_minutes=None,
        )
        out = QuotaFormatter()(quota)
        rich_cmp(out)

    def test_output_only_gpu(self, rich_cmp: RichCmp) -> None:
        quota = _Quota(
            total_running_jobs=None,
            total_gpu_run_time_minutes=9923,
            total_non_gpu_run_time_minutes=None,
        )
        out = QuotaFormatter()(quota)
        rich_cmp(out)

    def test_output_only_cpu(self, rich_cmp: RichCmp) -> None:
        quota = _Quota(
            total_running_jobs=None,
            total_non_gpu_run_time_minutes=3256,
            total_gpu_run_time_minutes=None,
        )
        out = QuotaFormatter()(quota)
        rich_cmp(out)

    def test_output_only_jobs(self, rich_cmp: RichCmp) -> None:
        quota = _Quota(
            total_running_jobs=10,
            total_non_gpu_run_time_minutes=None,
            total_gpu_run_time_minutes=None,
        )
        out = QuotaFormatter()(quota)
        rich_cmp(out)

    def test_output_zeroes(self, rich_cmp: RichCmp) -> None:
        quota = _Quota(
            total_running_jobs=0,
            total_gpu_run_time_minutes=0,
            total_non_gpu_run_time_minutes=0,
        )
        out = QuotaFormatter()(quota)
        rich_cmp(out)


class TestAliasesFormatter:
    async def test_output(self, root: Root, nmrc_path: Path, rich_cmp: RichCmp) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "lsl": {
                            "cmd": "storage ls -l",
                            "help": "Custom ls with long output.",
                        },
                        "user-cmd": {"exec": "script"},
                    }
                }
            )
        )
        lst = await list_aliases(root)
        out = AliasesFormatter()(lst)
        rich_cmp(out)



import inspect
import sys
from pathlib import Path
from typing import Callable, List

import pytest
import toml

from neuromation.cli.alias import find_alias, list_aliases
from neuromation.cli.root import Root

from .conftest import SysCapWithCode


_RunCli = Callable[[List[str]], SysCapWithCode]


@pytest.fixture
def script() -> str:
    script = Path(__file__).parent / "script.py"
    return (sys.executable + " " + str(script)).replace("\\", "/")


def test_unknown_command(run_cli: _RunCli) -> None:
    capture = run_cli(["unknown-cmd"])
    assert capture.code == 2
    assert "Usage:" in capture.err


class TestInternalAlias:
    def test_internal_alias_simple(self, run_cli: _RunCli, nmrc_path: Path) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(toml.dumps({"alias": {"user-cmd": {"cmd": "help ls"}}}))
        capture = run_cli(["user-cmd"])
        assert capture.code == 0
        assert "List directory contents" in capture.out

    def test_internal_alias_refers_to_unknown(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps({"alias": {"user-cmd": {"cmd": "unknown command"}}})
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 2
        assert (
            'Error: Alias user-cmd refers to unknown command "unknown"' in capture.err
        )

    def test_internal_alias_help(self, run_cli: _RunCli, nmrc_path: Path) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(toml.dumps({"alias": {"lsl": {"cmd": "storage ls -l"}}}))
        capture = run_cli(["lsl", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} lsl [OPTIONS]

            Alias for "pytest storage ls -l"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_internal_alias_help_custom_msg(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "lsl": {
                            "cmd": "storage ls -l",
                            "help": "Custom ls with long output.",
                        }
                    }
                }
            )
        )
        capture = run_cli(["lsl", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} lsl [OPTIONS]

            Alias for "pytest storage ls -l"

            Custom ls with long output.

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    async def test_internal_alias_short_help(self, root: Root, nmrc_path: Path) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(toml.dumps({"alias": {"lsl": {"cmd": "storage ls -l"}}}))
        cmd = await find_alias(root, "lsl")
        assert cmd is not None
        assert cmd.get_short_help_str() == "neuro storage ls -l"

    async def test_internal_alias_short_help_custom_msg(
        self, root: Root, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "lsl": {
                            "cmd": "storage ls -l",
                            "help": "Custom ls with long output.",
                        }
                    }
                }
            )
        )
        cmd = await find_alias(root, "lsl")
        assert cmd is not None
        assert cmd.get_short_help_str() == "Custom ls with long output."


class TestExternalAliasArgs:
    def test_external_alias_no_arg(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(toml.dumps({"alias": {"user-cmd": {"exec": script}}}))
        capture = run_cli(["user-cmd"])
        assert capture.code == 0
        assert "[]" == capture.out

    def test_external_alias_no_arg_help(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(toml.dumps({"alias": {"user-cmd": {"exec": "script"}}}))
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_external_alias_no_arg_help_custom_msg(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script", "help": "Custom help."}}}
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script"

            Custom help.

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    async def test_external_alias_short_help(self, root: Root, nmrc_path: Path) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(toml.dumps({"alias": {"user-cmd": {"exec": "script"}}}))
        cmd = await find_alias(root, "user-cmd")
        assert cmd is not None
        assert cmd.get_short_help_str() == "script"

    async def test_external_alias_short_help_custom_msg(
        self, root: Root, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script", "help": "Custom help."}}}
            )
        )
        cmd = await find_alias(root, "user-cmd")
        assert cmd is not None
        assert cmd.get_short_help_str() == "Custom help."

    def test_external_alias_arg(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": f"{script} {{arg}}", "args": "ARG"}}}
            )
        )
        capture = run_cli(["user-cmd", "argument"])
        assert capture.code == 0
        assert "['argument']" == capture.out

    def test_external_alias_arg_help(self, run_cli: _RunCli, nmrc_path: Path) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps({"alias": {"user-cmd": {"exec": "script {arg}", "args": "ARG"}}})
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS] ARG

            Alias for "script {{arg}}"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_external_alias_arg_help_fix_casing(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps({"alias": {"user-cmd": {"exec": "script {arg}", "args": "Arg"}}})
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS] ARG

            Alias for "script {{arg}}"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_external_alias_two_arg2(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": f"{script} {{arg1}} {{arg2}}",
                            "args": "ARG1 ARG2",
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "arg1", "arg2"])
        assert capture.code == 0
        assert "['arg1', 'arg2']" == capture.out

    def test_external_alias_two_arg2_help(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {arg1} {arg2}",
                            "args": "ARG1 ARG2",
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS] ARG1 ARG2

            Alias for "script {{arg1}} {{arg2}}"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_external_alias_optional_arg_provided(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": f"{script} {{arg}}", "args": "[ARG]"}}}
            )
        )
        capture = run_cli(["user-cmd", "argument"])
        assert capture.code == 0
        assert "['argument']" == capture.out

    def test_external_alias_optional_arg_missed(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": f"{script} {{arg}}", "args": "[ARG]"}}}
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 0
        assert "[]" == capture.out

    def test_external_alias_optional_arg_help(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "[ARG]"}}}
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS] [ARG]

            Alias for "script {{arg}}"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_external_alias_multiple_arg_provided(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": f"{script} {{arg}}", "args": "ARG..."}}}
            )
        )
        capture = run_cli(["user-cmd", "arg1", "arg2"])
        assert capture.code == 0
        assert "['arg1', 'arg2']" == capture.out

    def test_external_alias_multiple_arg_help(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "ARG..."}}}
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS] ARG...

            Alias for "script {{arg}}"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_external_alias_optional_multiple_arg_provided(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {"exec": f"{script} {{arg}}", "args": "[ARG]..."}
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "arg1", "arg2"])
        assert capture.code == 0
        assert "['arg1', 'arg2']" == capture.out

    def test_external_alias_optional_multiple_arg_missed(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {"exec": f"{script} {{arg}}", "args": "[ARG]..."}
                    }
                }
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 0
        assert "[]" == capture.out

    def test_external_alias_optional_multiple_arg_help(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "[ARG]..."}}}
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS] [ARG]...

            Alias for "script {{arg}}"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out

    def test_external_alias_three_args_regular_multiple_optional_help(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {arg1} {arg2} {arg3}",
                            "args": "ARG1 ARG2... [ARG3]",
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS] ARG1 ARG2... [ARG3]

            Alias for "script {{arg1}} {{arg2}} {{arg3}}"

            Options:
              --help  Show this message and exit.
        """
        )
        assert expected == capture.out


class TestExternalAliasOptions:
    def test_external_alias_option_flag_help_without_help_str(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {opt}", "options": ["--opt"]}}}
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script {{opt}}"

            Options:
              --help  Show this message and exit.
              --opt
        """
        )
        assert expected == capture.out

    def test_external_alias_option_flag_help_with_help_str(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {opt}",
                            "options": ["--opt  Option description."],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script {{opt}}"

            Options:
              --help  Show this message and exit.
              --opt   Option description.
        """
        )
        assert expected == capture.out

    def test_external_alias_option_flag_short_long_help_with_help_str(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {opt}",
                            "options": ["-o, --opt  Option description."],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script {{opt}}"

            Options:
              --help     Show this message and exit.
              -o, --opt  Option description.
        """
        )
        assert expected == capture.out

    def test_external_alias_option_short_long_help_with_help_str(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {opt}",
                            "options": ["-o, --opt VAL  Option description."],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script {{opt}}"

            Options:
              --help         Show this message and exit.
              -o, --opt VAL  Option description.
        """
        )
        assert expected == capture.out

    def test_external_alias_option_metaval_lowercased(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {opt}",
                            "options": ["-o, --opt val  Description."],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script {{opt}}"

            Options:
              --help         Show this message and exit.
              -o, --opt VAL  Description.
        """
        )
        assert expected == capture.out

    def test_external_alias_option_short_long_help_with_help_str_inversed_order(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {opt}",
                            "options": ["--opt, -o VAL  Option description."],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--help"])
        assert capture.code == 0
        prog_name = Path(sys.argv[0]).name
        expected = inspect.cleandoc(
            f"""\
            Usage: {prog_name} user-cmd [OPTIONS]

            Alias for "script {{opt}}"

            Options:
              --help         Show this message and exit.
              -o, --opt VAL  Option description.
        """
        )
        assert expected == capture.out

    def test_external_alias_option_call_flag_short(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": f"{script} {{opt}}",
                            "options": ["-o, --opt"],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "-o"])
        assert capture.code == 0, capture
        assert capture.out == "['--opt']"

    def test_external_alias_option_call_flag_long(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": f"{script} {{opt}}",
                            "options": ["-o, --opt"],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--opt"])
        assert capture.code == 0, capture
        assert capture.out == "['--opt']"

    def test_external_alias_option_call_flag_unset(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": f"{script} {{opt}}",
                            "options": ["-o, --opt"],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 0, capture
        assert capture.out == "[]"

    def test_external_alias_option_call_value(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": f"{script} {{opt}}",
                            "options": ["-o, --opt VAL"],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--opt", "arg"])
        assert capture.code == 0, capture
        assert capture.out == "['--opt', 'arg']"

    def test_external_alias_option_call_value_multiple(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": f"{script} {{opt}}",
                            "options": ["-o, --opt VAL"],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--opt", "arg1", "--opt", "arg2"])
        assert capture.code == 0, capture
        assert capture.out == "['--opt', 'arg1', '--opt', 'arg2']"

    def test_external_alias_option_call_flag_multiple(
        self, run_cli: _RunCli, nmrc_path: Path, script: str
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": f"{script} {{verbose}}",
                            "options": ["-v, --verbose"],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "-vvv"])
        assert capture.code == 0, capture
        assert capture.out == "['--verbose', '--verbose', '--verbose']"


def test_external_alias_exitcode(
    run_cli: _RunCli, nmrc_path: Path, script: str
) -> None:
    user_cfg = nmrc_path / "user.toml"
    user_cfg.write_text(
        toml.dumps(
            {
                "alias": {
                    "user-cmd": {
                        "exec": f"{script} {{exit}}",
                        "options": ["--exit CODE"],
                    }
                }
            }
        )
    )
    capture = run_cli(["user-cmd", "--exit=10"])
    assert capture.code == 10, capture
    assert capture.out == "['--exit', '10']"


class TestExternalAliasParseErrors:
    def test_external_alias_long_option_not_identifier(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {opt}", "options": ["--123"]}}}
            )
        )
        capture = run_cli(["user-cmd", "--opt", "arg"])
        assert capture.code == 70, capture
        assert capture.err.startswith("Cannot parse option --123")

    def test_external_alias_short_option_not_identifier(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {opt}", "options": ["-1"]}}}
            )
        )
        capture = run_cli(["user-cmd", "--opt", "arg"])
        assert capture.code == 70, capture
        assert capture.err.startswith("Cannot parse option -1")

    def test_external_alias_option_meta_not_identifier(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {"exec": "script {opt}", "options": ["--opt 123"]}
                    }
                }
            )
        )
        capture = run_cli(["user-cmd", "--opt", "arg"])
        assert capture.code == 70, capture
        assert capture.err.startswith("Cannot parse option --opt 123")

    def test_external_alias_empty_substitution(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(toml.dumps({"alias": {"user-cmd": {"exec": "script {}"}}}))
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith("Empty substitution is not allowed")

    def test_external_alias_uppercased_parameter(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps({"alias": {"user-cmd": {"exec": "script {PARAM}"}}})
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith("Parameter PARAM should be lowercased")

    def test_external_alias_invalid_parameter_name(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps({"alias": {"user-cmd": {"exec": "script {123}"}}})
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith("Parameter 123 is not a valid identifier")

    def test_external_alias_unknown_parameter(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {param}",
                            "options": ["-v, --verbose"],
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Unknown parameter param in "script {param}"')

    def test_external_alias_overlapped_args_and_options(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {
                    "alias": {
                        "user-cmd": {
                            "exec": "script {param}",
                            "options": ["--param"],
                            "args": "PARAM",
                        }
                    }
                }
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith(
            "The following names are present in both positional "
            "and optional arguments: param"
        )

    def test_external_alias_nested_args_brackets(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "[[ARG]]"}}}
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Nested brackets in "[[ARG]]"')

    def test_external_alias_missing_open_bracket(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "ARG]"}}}
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Missing open bracket in "ARG]"')

    def test_external_alias_missing_argument_inside_brackets(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps({"alias": {"user-cmd": {"exec": "script {arg}", "args": "[]"}}})
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Missing argument inside brackets in "[]"')

    def test_external_alias_ellipsis_should_follow_arg(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps({"alias": {"user-cmd": {"exec": "script {arg}", "args": "..."}}})
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith(
            'Ellipsis (...) should follow an argument in "..."'
        )

    def test_external_alias_ellipsis_inside_brackets(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "[ARG...]"}}}
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Ellipsis (...) inside of brackets in "[ARG...]"')

    def test_external_alias_successive_ellipsis(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "ARG......"}}}
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Successive ellipsis (...) in "ARG......"')

    def test_external_alias_missing_close_bracket1(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "[ARG"}}}
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Missing close bracket in "[ARG"')

    def test_external_alias_missing_close_bracket2(
        self, run_cli: _RunCli, nmrc_path: Path
    ) -> None:
        user_cfg = nmrc_path / "user.toml"
        user_cfg.write_text(
            toml.dumps(
                {"alias": {"user-cmd": {"exec": "script {arg}", "args": "[ARG1 ARG2"}}}
            )
        )
        capture = run_cli(["user-cmd"])
        assert capture.code == 70, capture
        assert capture.err.startswith('Missing close bracket in "[ARG1 ARG2"')


def test_external_alias_simplified(
    run_cli: _RunCli, nmrc_path: Path, script: str
) -> None:
    user_cfg = nmrc_path / "user.toml"
    user_cfg.write_text(
        toml.dumps(
            {
                "alias": {
                    "user-cmd": {
                        "exec": f"{script}",
                        "args": "[ARG]...",
                        "options": ["-o, --opt  Option"],
                    }
                }
            }
        )
    )
    capture = run_cli(["user-cmd", "-o", "arg"])
    assert capture.code == 0
    assert "['--opt', 'arg']" == capture.out


async def _test_list_aliases(root: Root) -> None:
    user_cfg = root.config_path / "user.toml"
    user_cfg.write_text(
        toml.dumps(
            {
                "alias": {
                    "lsl": {
                        "cmd": "storage ls -l",
                        "help": "Custom ls with long output.",
                    },
                    "user-cmd": {"exec": "script"},
                }
            }
        )
    )
    lst = await list_aliases(root)
    names = [cmd.name for cmd in lst]
    assert names == ["lsl", "user-cmd"]


async def test_list_aliases(root: Root) -> None:
    await _test_list_aliases(root)


async def test_list_aliases_no_logged_in(root_no_logged_in: Root) -> None:
    await _test_list_aliases(root_no_logged_in)


async def test_find_alias_no_logged_in(root_no_logged_in: Root) -> None:
    user_cfg = root_no_logged_in.config_path / "user.toml"
    user_cfg.write_text(
        toml.dumps(
            {
                "alias": {
                    "lsl": {
                        "cmd": "storage ls -l",
                        "help": "Custom ls with long output.",
                    },
                    "user-cmd": {"exec": "script"},
                }
            }
        )
    )
    cmd = await find_alias(root_no_logged_in, "lsl")
    assert cmd is not None
    assert cmd.get_short_help_str() == "Custom ls with long output."


async def test_find_alias_without_config(root_no_logged_in: Root) -> None:
    assert await find_alias(root_no_logged_in, "unknown-cmd") is None



import io
import json
import logging
import sys
from collections import namedtuple
from pathlib import Path
from typing import Any, Callable, List

import pytest
from yarl import URL

from neuromation.api import CONFIG_ENV_NAME, Config
from neuromation.cli.const import EX_OK
from neuromation.cli.docker_credential_helper import main as dch
from neuromation.cli.root import Root


SysCapWithCode = namedtuple("SysCapWithCode", ["out", "err", "code"])
log = logging.getLogger(__name__)


_RunCli = Callable[[List[str]], SysCapWithCode]


@pytest.fixture()
def config(root: Root) -> Config:
    return root.client.config


_RunDch = Callable[[List[str]], SysCapWithCode]


@pytest.fixture()
def run_dch(capfd: Any, monkeypatch: Any, tmp_path: Path, nmrc_path: Path) -> _RunDch:
    def _run_dch(arguments: List[str]) -> SysCapWithCode:

        log.info("Run 'docker-helper-neuro %s'", " ".join(arguments))
        code = EX_OK
        try:
            with monkeypatch.context() as ctx:
                ctx.setattr(sys, "argv", ["docker-credential-helper"] + arguments)
                ctx.setenv(CONFIG_ENV_NAME, str(nmrc_path))
                dch()
        except SystemExit as e:
            code = e.code
        out, err = capfd.readouterr()
        return SysCapWithCode(out.strip(), err.strip(), code)

    return _run_dch


class TestCli:
    def test_path_not_exists(self, run_cli: _RunCli, tmp_path: Path) -> None:
        path = tmp_path / "some" / "not-exists"
        json_path = path / "config.json"
        capture = run_cli(["config", "docker", "--docker-config", str(path)])
        assert not capture.err
        assert json_path.is_file()

    def test_path_is_not_dir(self, run_cli: _RunCli, tmp_path: Path) -> None:
        path = tmp_path / "file"
        with path.open("w") as file:
            file.write("text")
        captured = run_cli(["config", "docker", "--docker-config", str(path)])
        assert captured.code
        assert captured.err

    def test_path_from_env(
        self, run_cli: _RunCli, tmp_path: Path, monkeypatch: Any, config: Config
    ) -> None:
        json_path = tmp_path / "config.json"
        with json_path.open("w") as file:
            file.write("{}")
        monkeypatch.setenv("DOCKER_CONFIG", str(tmp_path))
        capture = run_cli(["config", "docker"])
        assert not capture.err
        assert json_path.is_file()
        with json_path.open("rb") as fp:
            payload = json.load(fp)
        registry = URL(config.clusters[config.cluster_name].registry_url).host
        assert payload["credHelpers"] == {registry: "neuro"}

    def test_new_file(self, run_cli: _RunCli, tmp_path: Path, config: Config) -> None:
        path = tmp_path / ".docker"
        json_path = path / "config.json"
        capture = run_cli(["config", "docker", "--docker-config", str(path)])
        assert not capture.err
        assert json_path.is_file()
        with json_path.open("rb") as fp:
            payload = json.load(fp)
        registry = URL(config.clusters[config.cluster_name].registry_url).host
        assert payload["credHelpers"] == {registry: "neuro"}

    def test_merge_file_without_helpers(
        self, run_cli: _RunCli, tmp_path: Path, config: Config
    ) -> None:
        path = tmp_path / ".docker"
        path.mkdir()
        json_path = path / "config.json"
        with json_path.open("w", encoding="utf-8") as fp:
            json.dump({"test": "value\u20ac"}, fp)
        capture = run_cli(["config", "docker", "--docker-config", str(path)])
        assert not capture.err
        assert json_path.is_file()
        with json_path.open("rb") as fp2:
            payload = json.load(fp2)
        registry = URL(config.clusters[config.cluster_name].registry_url).host
        assert payload["credHelpers"] == {registry: "neuro"}
        assert payload["test"] == "value\u20ac"

    def test_merge_file_with_existing_helpers(
        self, run_cli: _RunCli, tmp_path: Path, config: Config
    ) -> None:
        path = tmp_path / ".docker"
        path.mkdir()
        json_path = path / "config.json"
        with json_path.open("w", encoding="utf-8") as fp:
            json.dump(
                {"test": "value\u20ac", "credHelpers": {"some.com": "handler"}}, fp
            )
        capture = run_cli(["config", "docker", "--docker-config", str(path)])
        assert not capture.err
        assert json_path.is_file()
        with json_path.open("rb") as fp2:
            payload = json.load(fp2)
        registry = URL(config.clusters[config.cluster_name].registry_url).host
        assert payload["credHelpers"] == {registry: "neuro", "some.com": "handler"}
        assert payload["test"] == "value\u20ac"

    def test_success_output_message(
        self, run_cli: _RunCli, tmp_path: Path, config: Config
    ) -> None:
        path = tmp_path / ".docker"
        json_path = path / "config.json"
        capture = run_cli(["config", "docker", "--docker-config", str(path)])
        assert not capture.err
        assert str(json_path) in capture.out.replace("\n", "")  # strip rich wrapping
        assert config.clusters[config.cluster_name].registry_url.host in capture.out


class TestHelper:
    def test_no_params_use(self, run_dch: _RunDch) -> None:
        capture = run_dch([])
        assert capture.code != EX_OK

    def test_too_mach_params(self, run_dch: _RunDch) -> None:
        capture = run_dch(["one", "two"])
        assert capture.code != EX_OK

    def test_unknown_operation(self, run_dch: _RunDch) -> None:
        capture = run_dch(["ping"])
        assert capture.code != EX_OK

    def test_store_operation(self, run_dch: _RunDch) -> None:
        capture = run_dch(["store"])
        assert capture.code != EX_OK

    def test_get_operation(
        self, run_dch: _RunDch, monkeypatch: Any, config: Config, token: str
    ) -> None:
        registry = config.clusters[config.cluster_name].registry_url.host
        assert registry is not None
        monkeypatch.setattr("sys.stdin", io.StringIO(registry))
        capture = run_dch(["get"])
        assert capture.code == EX_OK
        payload = json.loads(capture.out)
        assert payload == {"Username": "token", "Secret": token}



import click
import pytest

from neuromation.cli.text_helper import StyledTextHelper


class TestStyledTextHelper:
    @pytest.mark.parametrize(
        "text, expected",
        [("Simple text", False), (click.style("Harder ", bold=True) + "text", True)],
    )
    def test_is_styled(self, text: str, expected: bool) -> None:
        assert StyledTextHelper.is_styled(text) is expected

    @pytest.mark.parametrize(
        "text, expected",
        [("Simple text", 11), (click.style("Harder ", bold=True) + "text", 11)],
    )
    def test_width(self, text: str, expected: int) -> None:
        assert StyledTextHelper.width(text) == expected

    @pytest.mark.parametrize(
        "text, width, expected",
        [
            # Not styled cases
            ("Simple text", 100, "Simple text"),
            ("Simple text", 6, "Simple"),
            # Styled cases
            (
                click.style("Harder ", bold=True) + "text",
                7,
                click.style("Harder ", bold=True),
            ),
            (
                click.style("Harder ", bold=True) + "text",
                9,
                click.style("Harder ", bold=True) + "te",
            ),
            (
                click.style("Harder ", bold=True) + "text",
                4,
                click.style("Hard", bold=True),
            ),
            (
                click.style("Very ", underline=True, bold=True, reset=False)
                + click.style("Hard", bold=True)
                + "text",
                4,
                click.style("Very", underline=True, bold=True),
            ),
            (
                click.style("Very ", underline=True, bold=True, reset=False)
                + click.style("Hard", bold=True)
                + "text",
                8,
                click.style("Very ", underline=True, bold=True, reset=False)
                + click.style("Har", bold=True),
            ),
        ],
    )
    def test_trim(self, text: str, width: int, expected: str) -> None:
        assert StyledTextHelper.trim(text, width) == expected

    @pytest.mark.parametrize(
        "text, width, expected",
        [
            # Not styled cases
            ("Simple text", 100, ["Simple text"]),
            ("Simple text", 6, ["Simple", "text"]),
            ("Simple text", 4, ["Simp", "le", "text"]),
            # Styled cases
            (
                click.style("Harder ", bold=True) + "text",
                100,
                [click.style("Harder ", bold=True) + "text"],
            ),
            (
                click.style("Harder ", bold=True) + "text",
                6,
                [click.style("Harder", bold=True), "text"],
            ),
            (
                click.style("Harder ", bold=True) + "text",
                4,
                [click.style("Hard", bold=True), click.style("er", bold=True), "text"],
            ),
            (
                click.style("Very ", underline=True, reset=False)
                + click.style("Hard", bold=True)
                + " text",
                4,
                [
                    click.style("Very", underline=True),
                    click.style("", underline=True, reset=False)
                    + click.style("Hard", bold=True),
                    "text",
                ],
            ),
        ],
    )
    def test_wrap(self, text: str, width: int, expected: str) -> None:
        assert StyledTextHelper.wrap(text, width) == expected



import re
from difflib import SequenceMatcher
from textwrap import fill, wrap
from typing import Iterator, List, Sequence

import click
from wcwidth import wcswidth


class StyledTextHelper:
    """Utilities to perform string operations on terminal styled text with better
    unicode support.

    NOTE: The helper does not recognise controll sequences other than SGR
        (Select Graphic Rendition) ones.
    """

    style_re = re.compile(r"(\033\[(?:\d|;)*m)")
    ansi_reset_all = "\033[0m"
    whitespace = "\t\n\x0b\x0c\r "

    @classmethod
    def is_styled(cls, text: str) -> bool:
        return "\033" in text

    # As we often do operations on headers or similar static text it makes sense to
    # cache
    @classmethod
    def unstyle(cls, text: str) -> str:
        return click.unstyle(text)

    @classmethod
    def width(cls, text: str) -> int:
        if cls.is_styled(text):
            return wcswidth(cls.unstyle(text))
        else:
            return wcswidth(text)

    @classmethod
    def ljust(cls, text: str, width: int) -> str:
        if cls.is_styled(text):
            return text.ljust(width + len(text) - len(cls.unstyle(text)))
        else:
            return text.ljust(width)

    @classmethod
    def rjust(cls, text: str, width: int) -> str:
        if cls.is_styled(text):
            return text.rjust(width + len(text) - len(cls.unstyle(text)))
        else:
            return text.rjust(width)

    @classmethod
    def center(cls, text: str, width: int) -> str:
        if cls.is_styled(text):
            return text.center(width + len(text) - len(cls.unstyle(text)))
        else:
            return text.center(width)

    @classmethod
    def trim(cls, text: str, width: int) -> str:
        if not cls.is_styled(text):
            return text[:width]

        result = []
        has_unclosed = False
        remaining = width
        for token in cls.style_re.split(text):
            if cls.style_re.match(token):
                result.append(token)
                # We will need to add a reset at the end if there are any non-closed
                # sequences
                has_unclosed = token != cls.ansi_reset_all
            else:
                if len(token) >= remaining:
                    result.append(token[:remaining])
                    break
                remaining -= len(token)
                result.append(token)

        if has_unclosed:
            result.append(cls.ansi_reset_all)

        return "".join(result)

    @classmethod
    def wrap(cls, text: str, width: int) -> Sequence[str]:
        if cls.is_styled(text):
            # Fast return if we don't need to wrap the text
            if cls.width(text) <= width:
                return [text]

            wrapped = fill(cls.unstyle(text), width)
            return list(cls._reapply_styles(text, wrapped))
        else:
            return wrap(text, width)

    @classmethod
    def _reapply_styles(cls, text: str, wrapped: str) -> Iterator[str]:
        """Iterate over wrapped text and original one and reapply styles from
        the original to wrapped
        """
        sm = SequenceMatcher(None, text, wrapped)

        result: List[str] = []

        for op, ti, tj, wi, wj in sm.get_opcodes():
            if op == "equal" or op == "insert":
                # For those 2 cases we just take the text wrap() generated for us
                result.append(wrapped[wi:wj])
            if op == "delete" or op == "replace":
                # In those cases we either replaced some whitespace's or need to apply
                # styles from original.

                # The most complicated situation would be if for some reason we had
                # several whitespaces in the styled case that were not replaced by wrap.
                # In this case we have to assume that the positioning of those style
                # codes matter (ex. we want to have 1 space before and after a word
                # underlined). Ex:
                #
                #   Plase select exactly \033[4m one \033[0m word
                #
                ws_buf = wrapped[wi:wj]
                for token in cls.style_re.split(text[ti:tj]):
                    if cls.style_re.match(token):
                        result.append(token)
                    else:
                        # We take the same amount of whitespace from result to preserve
                        # style opcode position
                        ws_len = len(token)
                        # NOTE: ws_buf can be empty at any point
                        result.append(ws_buf[:ws_len])
                        ws_buf = ws_buf[ws_len:]

        # Fix style sequencing on breaklines
        ansi_stack: List[str] = []
        ansi_reset_all = cls.ansi_reset_all
        for line in "".join(result).split("\n"):
            # Strip any reset styling at beginning. Ex: "\033[4m\033[0m word"
            prefix, _, remainder = line.partition(ansi_reset_all)
            if not cls.unstyle(prefix):
                line = remainder
                # We also assume we broke all sequnces from previous lines
                ansi_stack = []

            # Reopen any sequences that were opened before breakline
            if ansi_stack:
                line = "".join(ansi_stack) + line

            # Find sequnces not closed at the end of this line
            ansi_stack = []
            for token in cls.style_re.findall(line):
                if token == ansi_reset_all:
                    ansi_stack = []
                else:
                    # XXX: Add excluding op-codes handling too. Like bold (1) and
                    #      normal (22)
                    ansi_stack.append(token)

            # Break sequnces if we have any opened
            if ansi_stack:
                line += ansi_reset_all

            yield line
        return



import time
from typing import Any, List

import pytest
from yarl import URL

from neuromation.api import Action, FileStatus, FileStatusType
from neuromation.cli.formatters.storage import (
    BaseFilesFormatter,
    BSDAttributes,
    BSDPainter,
    FilesSorter,
    GnuIndicators,
    GnuPainter,
    LongFilesFormatter,
    NonePainter,
    SimpleFilesFormatter,
    VerticalColumnsFilesFormatter,
    get_painter,
)


class TestNonePainter:
    def test_simple(self, rich_cmp: Any) -> None:
        painter = NonePainter()
        file = FileStatus(
            "File1",
            2048,
            FileStatusType.FILE,
            int(time.mktime(time.strptime("2018-01-01 03:00:00", "%Y-%m-%d %H:%M:%S"))),
            Action.READ,
            uri=URL("storage://default/user/File1"),
        )
        rich_cmp(painter.paint(file.name, file.type))


class TestGnuPainter:
    def test_color_parsing_simple(self) -> None:
        painter = GnuPainter("rs=1;0;1")
        assert painter.color_indicator[GnuIndicators.RESET] == "1;0;1"

        painter = GnuPainter(":rs=1;0;1")
        assert painter.color_indicator[GnuIndicators.RESET] == "1;0;1"

        painter = GnuPainter("rs=1;0;1:")
        assert painter.color_indicator[GnuIndicators.RESET] == "1;0;1"

        painter = GnuPainter("rs=1;0;1:fi=32;42")
        assert painter.color_indicator[GnuIndicators.RESET] == "1;0;1"
        assert painter.color_indicator[GnuIndicators.FILE] == "32;42"

        painter = GnuPainter("rs=1;0;1:fi")
        assert painter.color_indicator[GnuIndicators.RESET] == "1;0;1"
        assert painter.color_indicator[GnuIndicators.FILE] == ""

        painter = GnuPainter("rs=1;0;1:fi=")
        assert painter.color_indicator[GnuIndicators.RESET] == "1;0;1"
        assert painter.color_indicator[GnuIndicators.FILE] == ""

    @pytest.mark.parametrize(
        "escaped,result",
        [
            ("\\a", "\a"),
            ("\\b", "\b"),
            ("\\e", chr(27)),
            ("\\f", "\f"),
            ("\\n", "\n"),
            ("\\r", "\r"),
            ("\\t", "\t"),
            ("\\v", "\v"),
            ("\\?", chr(127)),
            ("\\_", " "),
            ("a\\n", "a\n"),
            ("a\\tb", "a\tb"),
            ("a\\t\\rb", "a\t\rb"),
            ("a\\=b", "a=b"),
        ],
    )
    def test_color_parsing_escaped_simple(self, escaped: str, result: str) -> None:
        painter = GnuPainter("rs=" + escaped)
        assert painter.color_indicator[GnuIndicators.RESET] == result

        painter = GnuPainter(escaped + "=1;2")
        assert painter.color_ext_type[result] == "1;2"

        painter = GnuPainter(escaped + "=" + escaped)
        assert painter.color_ext_type[result] == result

    @pytest.mark.parametrize(
        "escaped,result",
        [
            ("\\7", chr(7)),
            ("\\8", "8"),
            ("\\10", chr(8)),
            ("a\\2", "a" + chr(2)),
            ("a\\2b", "a" + chr(2) + "b"),
        ],
    )
    def test_color_parsing_escaped_octal(self, escaped: str, result: str) -> None:
        painter = GnuPainter("rs=" + escaped)
        assert painter.color_indicator[GnuIndicators.RESET] == result

        painter = GnuPainter(escaped + "=1;2")
        assert painter.color_ext_type[result] == "1;2"

        painter = GnuPainter(escaped + "=" + escaped)
        assert painter.color_ext_type[result] == result

    @pytest.mark.parametrize(
        "escaped,result",
        [
            ("\\x7", chr(0x7)),
            ("\\x8", chr(0x8)),
            ("\\x10", chr(0x10)),
            ("\\XaA", chr(0xAA)),
            ("a\\x222", "a" + chr(0x22) + "2"),
            ("a\\x2z", "a" + chr(0x2) + "z"),
        ],
    )
    def test_color_parsing_escaped_hex(self, escaped: str, result: str) -> None:
        painter = GnuPainter("rs=" + escaped)
        assert painter.color_indicator[GnuIndicators.RESET] == result

        painter = GnuPainter(escaped + "=1;2")
        assert painter.color_ext_type[result] == "1;2"

        painter = GnuPainter(escaped + "=" + escaped)
        assert painter.color_ext_type[result] == result

    @pytest.mark.parametrize(
        "escaped,result",
        [
            ("^a", chr(1)),
            ("^?", chr(127)),
            ("^z", chr(26)),
            ("a^Z", "a" + chr(26)),
            ("a^Zb", "a" + chr(26) + "b"),
        ],
    )
    def test_color_parsing_carret(self, escaped: str, result: str) -> None:
        painter = GnuPainter("rs=" + escaped)
        assert painter.color_indicator[GnuIndicators.RESET] == result

        painter = GnuPainter(escaped + "=1;2")
        assert painter.color_ext_type[result] == "1;2"

        painter = GnuPainter(escaped + "=" + escaped)
        assert painter.color_ext_type[result] == result

    @pytest.mark.parametrize("escaped", [("^1"), ("^"), ("^" + chr(130))])
    def test_color_parsing_carret_incorrect(self, escaped: str) -> None:
        with pytest.raises(EnvironmentError):
            GnuPainter("rs=" + escaped)

        with pytest.raises(EnvironmentError):
            GnuPainter(escaped + "=1;2")

    @pytest.mark.parametrize(
        "ls_colors",
        [
            "di=32;41:fi=0;44:no=0;46",
            "di=32;41:no=0;46",
            "no=0;46",
            "*.text=0;46",
            "*.txt=0;46",
        ],
    )
    def test_coloring(self, rich_cmp: Any, ls_colors: str) -> None:
        file = FileStatus(
            "test.txt",
            1024,
            FileStatusType.FILE,
            int(time.mktime(time.strptime("2018-01-01 03:00:00", "%Y-%m-%d %H:%M:%S"))),
            Action.READ,
            uri=URL("storage://default/usertest.txt"),
        )
        folder = FileStatus(
            "tmp",
            0,
            FileStatusType.DIRECTORY,
            int(time.mktime(time.strptime("2018-01-01 03:00:00", "%Y-%m-%d %H:%M:%S"))),
            Action.WRITE,
            uri=URL("storage://default/usertmp"),
        )
        painter = GnuPainter(ls_colors)
        rich_cmp(painter.paint(file.name, file.type), index=0)
        rich_cmp(painter.paint(folder.name, folder.type), index=1)


class TestBSDPainter:
    def test_color_parsing(self) -> None:
        painter = BSDPainter("exfxcxdxbxegedabagacad")
        assert painter._colors[BSDAttributes.DIRECTORY] == "ex"

    @pytest.mark.parametrize(
        "ls_colors", ["exfxcxdxbxegedabagacad", "Eafxcxdxbxegedabagacad"]
    )
    def test_coloring(self, ls_colors: str, rich_cmp: Any) -> None:
        file = FileStatus(
            "test.txt",
            1024,
            FileStatusType.FILE,
            int(time.mktime(time.strptime("2018-01-01 03:00:00", "%Y-%m-%d %H:%M:%S"))),
            Action.READ,
            uri=URL("storage://default/usertest.txt"),
        )
        folder = FileStatus(
            "tmp",
            0,
            FileStatusType.DIRECTORY,
            int(time.mktime(time.strptime("2018-01-01 03:00:00", "%Y-%m-%d %H:%M:%S"))),
            Action.WRITE,
            uri=URL("storage://default/usertmp"),
        )
        painter = BSDPainter(ls_colors)
        rich_cmp(painter.paint(file.name, file.type), index=0)
        rich_cmp(painter.paint(folder.name, folder.type), index=1)


class TestPainterFactory:
    def test_detection(self, monkeypatch: Any) -> None:
        monkeypatch.setenv("LS_COLORS", "")
        monkeypatch.setenv("LSCOLORS", "")
        painter = get_painter(True)
        assert isinstance(painter, NonePainter)

        monkeypatch.setenv("LSCOLORS", "exfxcxdxbxegedabagacad")
        monkeypatch.setenv("LS_COLORS", "di=32;41:fi=0;44:no=0;46")
        painter_without_color = get_painter(False)
        painter_with_color = get_painter(True)
        assert isinstance(painter_without_color, NonePainter)
        assert not isinstance(painter_with_color, NonePainter)

        monkeypatch.setenv("LSCOLORS", "")
        monkeypatch.setenv("LS_COLORS", "di=32;41:fi=0;44:no=0;46")
        painter = get_painter(True)
        assert isinstance(painter, GnuPainter)

        monkeypatch.setenv("LSCOLORS", "exfxcxdxbxegedabagacad")
        monkeypatch.setenv("LS_COLORS", "")
        painter = get_painter(True)
        assert isinstance(painter, BSDPainter)


class TestFilesFormatter:

    files = [
        FileStatus(
            "File1",
            2048,
            FileStatusType.FILE,
            int(time.mktime(time.strptime("2018-01-01 03:00:00", "%Y-%m-%d %H:%M:%S"))),
            Action.READ,
            uri=URL("storage://default/userFile1"),
        ),
        FileStatus(
            "File2",
            1024,
            FileStatusType.FILE,
            int(time.mktime(time.strptime("2018-10-10 13:10:10", "%Y-%m-%d %H:%M:%S"))),
            Action.READ,
            uri=URL("storage://default/userFile2"),
        ),
        FileStatus(
            "File3 with space",
            1_024_001,
            FileStatusType.FILE,
            int(time.mktime(time.strptime("2019-02-02 05:02:02", "%Y-%m-%d %H:%M:%S"))),
            Action.READ,
            uri=URL("storage://default/userFile 3 with space"),
        ),
    ]
    folders = [
        FileStatus(
            "Folder1",
            0,
            FileStatusType.DIRECTORY,
            int(time.mktime(time.strptime("2017-03-03 06:03:03", "%Y-%m-%d %H:%M:%S"))),
            Action.MANAGE,
            uri=URL("storage://default/userFolder11"),
        ),
        FileStatus(
            "1Folder with space",
            0,
            FileStatusType.DIRECTORY,
            int(time.mktime(time.strptime("2017-03-03 06:03:02", "%Y-%m-%d %H:%M:%S"))),
            Action.MANAGE,
            uri=URL("storage://default/user1Folder with space"),
        ),
    ]
    files_and_folders = files + folders

    @pytest.mark.parametrize(
        "formatter",
        [
            (SimpleFilesFormatter(color=False)),
            (VerticalColumnsFilesFormatter(width=100, color=False)),
            (LongFilesFormatter(human_readable=False, color=False)),
        ],
    )
    def test_formatter_with_files_and_folders(
        self, formatter: BaseFilesFormatter, rich_cmp: Any
    ) -> None:
        rich_cmp(formatter(self.files_and_folders))

    @pytest.mark.parametrize(
        "formatter",
        [
            (SimpleFilesFormatter(color=False)),
            (VerticalColumnsFilesFormatter(width=100, color=False)),
            (LongFilesFormatter(human_readable=False, color=False)),
        ],
    )
    def test_formatter_with_empty_files(
        self, formatter: BaseFilesFormatter, rich_cmp: Any
    ) -> None:
        files: List[FileStatus] = []
        rich_cmp(formatter(files))

    def test_sorter(self) -> None:
        sorter = FilesSorter.NAME
        files = sorted(self.files_and_folders, key=sorter.key())
        assert files == [
            self.folders[1],
            self.files[0],
            self.files[1],
            self.files[2],
            self.folders[0],
        ]

        sorter = FilesSorter.SIZE
        files = sorted(self.files_and_folders, key=sorter.key())
        assert files[2:5] == [self.files[1], self.files[0], self.files[2]]

        sorter = FilesSorter.TIME
        files = sorted(self.files_and_folders, key=sorter.key())
        assert files == [
            self.folders[1],
            self.folders[0],
            self.files[0],
            self.files[1],
            self.files[2],
        ]



import json
from typing import Any, Callable, List, Mapping
from unittest import mock

from neuromation.api.admin import _Admin, _ClusterUser, _ClusterUserRoleType
from neuromation.api.server_cfg import Preset

from .conftest import SysCapWithCode


_RunCli = Callable[[List[str]], SysCapWithCode]


def test_add_cluster_user_print_result(run_cli: _RunCli) -> None:
    with mock.patch.object(_Admin, "add_cluster_user") as mocked:

        async def add_cluster_user(
            cluster_name: str, user_name: str, role: str
        ) -> _ClusterUser:
            # NOTE: We return a different role to check that we print it to user
            return _ClusterUser(user_name, _ClusterUserRoleType.MANAGER)

        mocked.side_effect = add_cluster_user
        capture = run_cli(["admin", "add-cluster-user", "default", "ivan", "admin"])
        assert not capture.err
        assert capture.out == "Added ivan to cluster default as manager"

        # Same with quiet mode
        mocked.side_effect = add_cluster_user
        capture = run_cli(
            ["-q", "admin", "add-cluster-user", "default", "ivan", "admin"]
        )
        assert not capture.err
        assert not capture.out


def test_remove_cluster_user_print_result(run_cli: _RunCli) -> None:
    with mock.patch.object(_Admin, "remove_cluster_user") as mocked:

        async def remove_cluster_user(cluster_name: str, user_name: str) -> None:
            return

        mocked.side_effect = remove_cluster_user
        capture = run_cli(["admin", "remove-cluster-user", "default", "ivan"])
        assert not capture.err
        assert capture.out == "Removed ivan from cluster default"

        # Same with quiet mode
        mocked.side_effect = remove_cluster_user
        capture = run_cli(["-q", "admin", "remove-cluster-user", "default", "ivan"])
        assert not capture.err
        assert not capture.out


def test_show_cluster_config_options(run_cli: _RunCli) -> None:
    with mock.patch.object(_Admin, "get_cloud_provider_options") as mocked:
        sample_data = {"foo": "bar", "baz": {"t2": 1, "t1": 2}}

        async def get_cloud_provider_options(
            cloud_provider_name: str,
        ) -> Mapping[str, Any]:
            assert cloud_provider_name == "aws"
            return sample_data

        mocked.side_effect = get_cloud_provider_options
        capture = run_cli(["admin", "show-cluster-options", "--type", "aws"])
        assert not capture.err

        assert json.loads(capture.out) == sample_data


def test_update_resource_preset(run_cli: _RunCli) -> None:
    with mock.patch.object(_Admin, "update_cluster_resource_presets") as mocked:

        async def update_cluster_resource_presets(
            cluster_name: str, presets: Mapping[str, Preset]
        ) -> None:
            assert cluster_name == "default"
            assert "cpu-micro" in presets
            assert presets["cpu-micro"] == Preset(
                cpu=0.1,
                memory_mb=100,
                gpu=1,
                gpu_model="nvidia-tesla-k80",
                tpu_type="v2-8",
                tpu_software_version="1.14",
                is_preemptible=True,
                is_preemptible_node_required=True,
            )

        mocked.side_effect = update_cluster_resource_presets
        capture = run_cli(
            [
                "admin",
                "update-resource-preset",
                "default",
                "cpu-micro",
                "-c",
                "0.1",
                "-m",
                "100M",
                "-g",
                "1",
                "--gpu-model",
                "nvidia-tesla-k80",
                "--tpu-type",
                "v2-8",
                "--tpu-sw-version",
                "1.14",
                "-p",
                "--preemptible-node",
            ]
        )
        assert capture.code == 0


def test_update_resource_preset_print_result(run_cli: _RunCli) -> None:
    with mock.patch.object(_Admin, "update_cluster_resource_presets") as mocked:

        async def update_cluster_resource_presets(
            cluster_name: str, presets: Mapping[str, Preset]
        ) -> None:
            pass

        mocked.side_effect = update_cluster_resource_presets
        capture = run_cli(["admin", "update-resource-preset", "default", "cpu-micro"])
        assert not capture.err
        assert capture.out == "Updated resource preset cpu-micro in cluster default"

        # Same with quiet mode
        capture = run_cli(
            ["-q", "admin", "update-resource-preset", "default", "cpu-micro"]
        )
        assert not capture.err
        assert not capture.out


def test_remove_resource_preset_print_result(run_cli: _RunCli) -> None:
    with mock.patch.object(_Admin, "update_cluster_resource_presets") as mocked:

        async def update_cluster_resource_presets(
            cluster_name: str, presets: Mapping[str, Preset]
        ) -> None:
            pass

        mocked.side_effect = update_cluster_resource_presets
        capture = run_cli(["admin", "remove-resource-preset", "default", "cpu-small"])
        assert not capture.err
        assert capture.out == "Removed resource preset cpu-small from cluster default"

        # Same with quiet mode
        capture = run_cli(
            ["-q", "admin", "remove-resource-preset", "default", "cpu-small"]
        )
        assert not capture.err
        assert not capture.out


def test_remove_resource_preset_not_exists(run_cli: _RunCli) -> None:
    with mock.patch.object(_Admin, "update_cluster_resource_presets") as mocked:

        async def update_cluster_resource_presets(
            cluster_name: str, presets: Mapping[str, Preset]
        ) -> None:
            pass

        mocked.side_effect = update_cluster_resource_presets
        capture = run_cli(["admin", "remove-resource-preset", "default", "unknown"])
        assert capture.code
        assert "Preset 'unknown' not found" in capture.err



import sys
from itertools import product
from pathlib import Path
from typing import Any, Callable, Iterator

import pytest
from yarl import URL

from neuromation.api import (
    StorageProgressComplete,
    StorageProgressDelete,
    StorageProgressEnterDir,
    StorageProgressFail,
    StorageProgressLeaveDir,
    StorageProgressStart,
    StorageProgressStep,
)
from neuromation.cli.formatters.storage import (
    DeleteProgress,
    StreamProgress,
    TTYProgress,
    create_storage_progress,
    format_url,
)
from neuromation.cli.root import Root
from tests.cli.conftest import NewConsole


class TimeCtl:
    def __init__(self) -> None:
        self._current: float = 0

    def tick(self, delta: float = 1.0) -> None:
        self._current += delta

    def get_time(self) -> float:
        return self._current


@pytest.fixture()
def time_ctl() -> TimeCtl:
    return TimeCtl()


def test_format_url_storage() -> None:
    u = URL("storage://asvetlov/folder")
    assert format_url(u) == "storage://asvetlov/folder"


def test_format_url_file() -> None:
    u = URL("file:///asvetlov/folder")
    if sys.platform == "win32":
        expected = "\\asvetlov\\folder"
    else:
        expected = "/asvetlov/folder"
    assert format_url(u) == expected


_MakeRoot = Callable[[bool, bool, bool], Root]


@pytest.fixture
def make_root(new_console: NewConsole) -> Iterator[_MakeRoot]:
    root = None

    def make(color: bool, tty: bool, verbose: bool) -> Root:
        nonlocal root
        root = Root(
            color,
            tty,
            True,
            60,
            Path("~/.neuro"),
            verbosity=int(verbose),
            trace=False,
            trace_hide_token=True,
            command_path="",
            command_params=[],
            skip_gmp_stats=True,
            show_traceback=False,
        )
        root.console = new_console(tty=tty, color=color)
        root.err_console = new_console(tty=tty, color=color)
        return root

    yield make
    if root is not None:
        root.close()


def test_progress_factory_none(make_root: _MakeRoot) -> None:
    progress = create_storage_progress(make_root(False, False, False), False)
    assert isinstance(progress, StreamProgress)
    progress.end()


def test_progress_factory_verbose(make_root: _MakeRoot) -> None:
    progress = create_storage_progress(make_root(False, False, False), False)
    assert isinstance(progress, StreamProgress)
    progress.end()


def test_progress_factory_percent(make_root: _MakeRoot) -> None:
    progress = create_storage_progress(make_root(False, False, False), True)
    assert isinstance(progress, TTYProgress)
    progress.end()


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows does not supports UNIX-like permissions",
)
@pytest.mark.parametrize(
    "color,tty,verbose,show_progress",
    product((False, True), repeat=4),
)
def test_progress(
    color: bool,
    tty: bool,
    verbose: bool,
    show_progress: bool,
    make_root: _MakeRoot,
    rich_cmp: Any,
    time_ctl: TimeCtl,
) -> None:
    root = make_root(color, tty, verbose)
    report = create_storage_progress(
        root, show_progress, get_time=time_ctl.get_time, auto_refresh=False
    )
    src = URL("file:///abc")
    dst = URL("storage:xyz")

    with report.begin(src, dst):
        rich_cmp(root.console, index=0)

        time_ctl.tick()
        report.enter(StorageProgressEnterDir(src, dst))
        rich_cmp(root.console, index=1)

        time_ctl.tick()
        report.start(StorageProgressStart(src, dst, 600))
        rich_cmp(root.console, index=2)

        time_ctl.tick()
        report.step(StorageProgressStep(src, dst, 300, 600))
        rich_cmp(root.console, index=3)

        time_ctl.tick()
        report.step(StorageProgressStep(src, dst, 400, 600))
        rich_cmp(root.console, index=4)

        time_ctl.tick()
        report.complete(StorageProgressComplete(src, dst, 600))
        rich_cmp(root.console, index=5)

        time_ctl.tick()
        report.leave(StorageProgressLeaveDir(src, dst))
        rich_cmp(root.console, index=6)
    rich_cmp(root.console, index=7)


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows does not supports UNIX-like permissions",
)
@pytest.mark.parametrize(
    "color,tty,verbose,show_progress",
    product((False, True), repeat=4),
)
def test_fail(
    color: bool,
    tty: bool,
    verbose: bool,
    show_progress: bool,
    rich_cmp: Any,
    make_root: _MakeRoot,
    time_ctl: TimeCtl,
) -> None:
    root = make_root(color, tty, verbose)
    report = create_storage_progress(
        root, show_progress, get_time=time_ctl.get_time, auto_refresh=False
    )
    src = URL("file:///abc")
    dst = URL("storage:xyz")
    with report.begin(src, dst):
        rich_cmp(root.console, index=0)

        report.fail(StorageProgressFail(src, dst, "error"))
        rich_cmp(root.err_console, index=1)
    rich_cmp(root.console, index=2)


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows does not supports UNIX-like permissions",
)
@pytest.mark.parametrize(
    "color,tty,verbose,show_progress",
    product((False, True), repeat=4),
)
def test_nested(
    color: bool,
    tty: bool,
    verbose: bool,
    show_progress: bool,
    rich_cmp: Any,
    make_root: _MakeRoot,
    time_ctl: TimeCtl,
) -> None:
    root = make_root(color, tty, verbose)
    report = create_storage_progress(
        root, show_progress, get_time=time_ctl.get_time, auto_refresh=False
    )
    src = URL("file:///abc")
    dst = URL("storage:xyz")
    src_f = URL("file:///abc/file.txt")
    dst_f = URL("storage:xyz/file.txt")
    src2 = URL("file:///abc/cde")
    dst2 = URL("storage:xyz/cde")
    src2_f = URL("file:///abc/cde/file.txt")
    dst2_f = URL("storage:xyz/cde/file.txt")

    with report.begin(src, dst):
        rich_cmp(root.console, index=0)

        report.enter(StorageProgressEnterDir(src, dst))
        rich_cmp(root.console, index=1)

        time_ctl.tick()
        report.start(StorageProgressStart(src_f, dst_f, 600))
        rich_cmp(root.console, index=2)

        time_ctl.tick()
        report.step(StorageProgressStep(src_f, dst_f, 300, 600))
        rich_cmp(root.console, index=3)

        time_ctl.tick()
        report.step(StorageProgressStep(src_f, dst_f, 400, 600))
        rich_cmp(root.console, index=4)

        time_ctl.tick()
        report.complete(StorageProgressComplete(src_f, dst_f, 600))
        rich_cmp(root.console, index=5)

        time_ctl.tick()
        report.enter(StorageProgressEnterDir(src2, dst2))
        rich_cmp(root.console, index=6)

        time_ctl.tick()
        report.start(StorageProgressStart(src2_f, dst2_f, 800))
        rich_cmp(root.console, index=7)

        time_ctl.tick()
        report.step(StorageProgressStep(src2_f, dst2_f, 300, 800))
        rich_cmp(root.console, index=8)

        time_ctl.tick()
        report.complete(StorageProgressComplete(src2_f, dst_f, 800))
        rich_cmp(root.console, index=9)

        time_ctl.tick()
        report.leave(StorageProgressLeaveDir(src2, dst2))
        rich_cmp(root.console, index=10)

        time_ctl.tick()
        report.leave(StorageProgressLeaveDir(src, dst))
        rich_cmp(root.console, index=11)
    rich_cmp(root.console, index=12)


@pytest.mark.parametrize(
    "color,tty,verbose",
    product((False, True), repeat=3),
)
def test_delete_progress(
    color: bool,
    tty: bool,
    verbose: bool,
    rich_cmp: Any,
    make_root: _MakeRoot,
    time_ctl: TimeCtl,
) -> None:
    root = make_root(color, tty, verbose)
    report = DeleteProgress(root)
    url_file = URL("storage:/abc/foo")
    url_dir = URL("storage:/abc")

    report.delete(StorageProgressDelete(url_file, is_dir=False))
    rich_cmp(root.console, index=0)

    report.delete(StorageProgressDelete(url_dir, is_dir=True))
    rich_cmp(root.console, index=1)



import asyncio
import contextlib
import logging
import os
import re
import sys
from dataclasses import dataclass, field
from http.cookies import Morsel  # noqa
from pathlib import Path
from types import SimpleNamespace
from typing import (
    Any,
    Awaitable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
)

import aiohttp
import click
from rich.console import Console, PagerContext
from rich.pager import Pager

from neuromation.api import Client, ConfigError, Factory, gen_trace_id
from neuromation.api.config import _ConfigData, load_user_config

from .asyncio_utils import Runner


log = logging.getLogger(__name__)


TEXT_TYPE = ("application/json", "text", "application/x-www-form-urlencoded")

HEADER_TOKEN_PATTERN = re.compile(
    r"(Bearer|Basic|Digest|Mutual)\s+(?P<token>[^ ]+\.[^ ]+\.[^ ]+)"
)


_T = TypeVar("_T")


class MaybePager(Pager):
    """Uses the pager installed on the system."""

    def __init__(self, console: Console) -> None:
        self._console = console
        self._limit = console.size[1] * 2 / 3

    def show(self, content: str) -> None:
        """Use the same pager used by pydoc."""
        if self._console.is_terminal and len(content.splitlines()) > self._limit:
            # Enforce ANSI sequence handling (colors etc.)
            os.environ["LESS"] = "-R"
            click.echo_via_pager(content)
        else:
            print(content, end="")


@dataclass
class Root:
    color: bool
    tty: bool
    disable_pypi_version_check: bool
    network_timeout: float
    config_path: Path
    trace: bool
    verbosity: int
    trace_hide_token: bool
    command_path: str
    command_params: List[Dict[str, Optional[str]]]
    skip_gmp_stats: bool
    show_traceback: bool

    _client: Optional[Client] = None
    _factory: Optional[Factory] = None
    _runner: Runner = field(init=False)
    console: Console = field(init=False)

    def __post_init__(self) -> None:
        self._runner = Runner(debug=self.verbosity >= 2)
        self._runner.__enter__()
        self.console = Console(
            color_system="auto" if self.color else None,
            force_terminal=self.tty,
            highlight=False,
            log_path=False,
        )
        if not self.console.is_terminal or self.console.is_dumb_terminal:
            # resize with wider width to prevent wrapping/cropping
            self.console = Console(
                color_system="auto" if self.color else None,
                force_terminal=self.tty,
                highlight=False,
                log_path=False,
                width=2048,
            )

        self.err_console = Console(
            file=sys.stderr,
            color_system="auto" if self.color else None,
            force_terminal=self.tty,
            highlight=False,
            log_path=False,
        )
        if not self.err_console.is_terminal or self.err_console.is_dumb_terminal:
            # resize with wider width to prevent wrapping/cropping
            self.err_console = Console(
                file=sys.stderr,
                color_system="auto" if self.color else None,
                force_terminal=self.tty,
                highlight=False,
                log_path=False,
                width=2048,
            )

    def close(self) -> None:
        if self._client is not None:
            self.run(self._client.close())

        try:
            # Suppress prints unhandled exceptions
            # on event loop closing
            sys.stderr = None  # type: ignore
            self._runner.__exit__(*sys.exc_info())
        finally:
            sys.stderr = sys.__stderr__

    def run(self, main: Awaitable[_T]) -> _T:
        return self._runner.run(main)

    @property
    def _config(self) -> _ConfigData:
        assert self._client is not None
        return self._client.config._config_data

    @property
    def quiet(self) -> bool:
        return self.verbosity < 0

    @property
    def terminal_size(self) -> Tuple[int, int]:
        return self.console.size

    @property
    def timeout(self) -> aiohttp.ClientTimeout:
        return aiohttp.ClientTimeout(
            None, None, self.network_timeout, self.network_timeout
        )

    @property
    def client(self) -> Client:
        assert self._client is not None
        return self._client

    @property
    def factory(self) -> Factory:
        if self._factory is None:
            trace_configs: Optional[List[aiohttp.TraceConfig]]
            if self.trace:
                trace_configs = [self._create_trace_config()]
            else:
                trace_configs = None
            self._factory = Factory(
                path=self.config_path,
                trace_configs=trace_configs,
                trace_id=gen_trace_id(),
            )
        return self._factory

    async def init_client(self) -> Client:
        if self._client is not None:
            return self._client
        client = await self.factory.get(timeout=self.timeout)

        self._client = client
        return self._client

    async def get_user_config(self) -> Mapping[str, Any]:
        try:
            client = await self.init_client()
        except ConfigError:
            return load_user_config(self.config_path.expanduser())
        else:
            return await client.config.get_user_config()

    def _create_trace_config(self) -> aiohttp.TraceConfig:
        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_start.append(self._on_request_start)
        trace_config.on_request_chunk_sent.append(self._on_request_chunk_sent)
        trace_config.on_request_end.append(self._on_request_end)
        trace_config.on_response_chunk_received.append(self._on_response_chunk_received)
        return trace_config

    def _print_debug(self, lines: List[str]) -> None:
        txt = "\n".join(click.style(line, dim=True) for line in lines)
        click.echo(txt, err=True)

    def _process_chunk(self, chunk: bytes, printable: bool) -> List[str]:
        if not chunk:
            return []
        if printable:
            return chunk.decode(errors="replace").split("\n")
        else:
            return [f"[binary {len(chunk)} bytes]"]

    async def _on_request_start(
        self,
        session: aiohttp.ClientSession,
        context: SimpleNamespace,
        data: aiohttp.TraceRequestStartParams,
    ) -> None:
        path = data.url.raw_path
        if data.url.raw_query_string:
            path += "?" + data.url.raw_query_string
        lines = [f"> {data.method} {path} HTTP/1.1"]
        for key, val in data.headers.items():
            if self.trace_hide_token:
                val = self._sanitize_header_value(val)
            lines.append(f"> {key}: {val}")
        lines.append("> ")
        self._print_debug(lines)

        content_type = data.headers.get("Content-Type", "")
        context.request_printable = content_type.startswith(TEXT_TYPE)

    async def _on_request_chunk_sent(
        self,
        session: aiohttp.ClientSession,
        context: SimpleNamespace,
        data: aiohttp.TraceRequestChunkSentParams,
    ) -> None:
        chunk = data.chunk
        lines = [
            "> " + line
            for line in self._process_chunk(chunk, context.request_printable)
        ]
        self._print_debug(lines)

    async def _on_request_end(
        self,
        session: aiohttp.ClientSession,
        context: SimpleNamespace,
        data: aiohttp.TraceRequestEndParams,
    ) -> None:
        lines = [f"< HTTP/1.1 {data.response.status} {data.response.reason}"]
        for key, val in data.response.headers.items():
            lines.append(f"< {key}: {val}")
        self._print_debug(lines)

        content_type = data.response.headers.get("Content-Type", "")
        context.response_printable = content_type.startswith(TEXT_TYPE)

    async def _on_response_chunk_received(
        self,
        session: aiohttp.ClientSession,
        context: SimpleNamespace,
        data: aiohttp.TraceResponseChunkReceivedParams,
    ) -> None:
        chunk = data.chunk
        lines = [
            "< " + line
            for line in self._process_chunk(chunk, context.response_printable)
        ]
        self._print_debug(lines)

    def _sanitize_header_value(self, text: str) -> str:
        for token in self._find_all_tokens(text):
            token_safe = self._sanitize_token(token)
            text = text.replace(token, token_safe)
        return text

    def _sanitize_token(self, token: str) -> str:
        tail_len: int = 5
        # at least a third part of the token should be hidden
        if tail_len >= len(token) // 3:
            return f"<hidden {len(token)} chars>"
        hidden = f"<hidden {len(token) - tail_len * 2} chars>"
        return token[:tail_len] + hidden + token[-tail_len:]

    def _find_all_tokens(self, text: str) -> Iterator[str]:
        for match in HEADER_TOKEN_PATTERN.finditer(text):
            yield match.group("token")

    async def cancel_with_logging(self, task: "asyncio.Task[Any]") -> None:
        if not task.done():
            task.cancel()
        try:
            with contextlib.suppress(asyncio.CancelledError):
                await task
        except Exception as exc:
            if self.show_traceback:
                log.exception(str(exc), stack_info=True)
            else:
                log.error(str(exc))

    def soft_reset_tty(self) -> None:
        if self.tty:
            # Soft reset the terminal.
            # For example, Midnight Commander often leaves
            # scrolling margins (DECSTBM) aligned only
            # to a part of the screen size
            sys.stdout.write("\x1b[!p")
            sys.stdout.flush()

    def pager(self) -> PagerContext:
        return self.console.pager(MaybePager(self.console), styles=True, links=True)

    def print(self, *objects: Any, err: bool = False, **kwargs: Any) -> None:
        if err:
            self.err_console.print(*objects, **kwargs)
        else:
            self.console.print(*objects, **kwargs)



import uuid
from pathlib import Path
from typing import Any

import pytest

from tests.e2e.conftest import Helper


@pytest.fixture
def secret_name() -> str:
    return "secret" + str(uuid.uuid4()).replace("-", "")[:10]


@pytest.mark.e2e
def test_create_list_delete(helper: Helper, secret_name: str) -> None:
    cap = helper.run_cli(["secret", "ls"])
    assert cap.err == ""
    assert secret_name not in cap.out

    cap = helper.run_cli(["secret", "add", secret_name, "value"])
    assert cap.err == ""

    cap = helper.run_cli(["secret", "ls"])
    assert cap.err == ""
    assert secret_name in cap.out

    cap = helper.run_cli(["secret", "rm", secret_name])
    assert cap.err == ""

    cap = helper.run_cli(["secret", "ls"])
    assert cap.err == ""
    assert secret_name not in cap.out


@pytest.mark.e2e
def test_create_from_file_list_delete(
    request: Any, helper: Helper, secret_name: str
) -> None:
    cap = helper.run_cli(["secret", "ls"])
    assert cap.err == ""
    assert secret_name not in cap.out

    secret_path = Path(f"~/test-secret-file-{uuid.uuid4()}")
    request.addfinalizer(secret_path.expanduser().unlink)
    secret_path.expanduser().write_bytes(b"value\xff\x00")
    cap = helper.run_cli(["secret", "add", secret_name, f"@{secret_path}"])
    assert cap.err == ""

    cap = helper.run_cli(["secret", "ls"])
    assert cap.err == ""
    assert secret_name in cap.out

    cap = helper.run_cli(["secret", "rm", secret_name])
    assert cap.err == ""

    cap = helper.run_cli(["secret", "ls"])
    assert cap.err == ""
    assert secret_name not in cap.out
